import asyncio
import logging
import socket
from typing import Any

from redis.asyncio import Redis as _Redis
from redis.asyncio.client import Pipeline
from redis.exceptions import NoScriptError
from redis.exceptions import RedisError as RedisConnectionError

from cashews.exceptions import CacheBackendInteractionError

logger = logging.getLogger(__name__)


class Redis(_Redis):
    async def execute_command(self, command, *args: Any, **kwargs: Any):
        try:
            return await super().execute_command(command, *args, **kwargs)
        except NoScriptError:
            # used by register_script functionality
            # if we do not reraise it, than a Script wrapper will not work as expect
            raise
        except (
            RedisConnectionError,
            socket.gaierror,
            OSError,
            asyncio.TimeoutError,
        ) as exp:
            raise CacheBackendInteractionError() from exp


class SafeRedis(_Redis):
    async def execute_command(self, command, *args: Any, **kwargs: Any):
        try:
            return await super().execute_command(command, *args, **kwargs)
        except NoScriptError:
            # used by register_script functionality
            # if we do not reraise it, than a Script wrapper will not work as expect
            raise
        except (
            RedisConnectionError,
            socket.gaierror,
            OSError,
            asyncio.TimeoutError,
        ) as exp:
            if command.lower() == "ping":
                raise CacheBackendInteractionError() from exp
            logger.error("redis: can not execute command: %s", command, exc_info=True)
            if command.lower() in ["unlink", "del", "memory", "ttl"]:
                return 0
            if command.lower() == "scan":
                return [0, []]
            return None

    async def initialize(self):
        try:
            return await super().initialize()
        except (RedisConnectionError, socket.gaierror, OSError, asyncio.TimeoutError):
            logger.error("redis: can not initialize cache", exc_info=True)
            return self

    __aenter__ = initialize


class SafePipeline(Pipeline):
    async def execute(self, raise_on_error=False):
        try:
            await super().execute(raise_on_error)
        except RedisConnectionError:
            logger.error("redis: can not execute pipeline", exc_info=True)



import asyncio

import pytest

from cashews.exceptions import CircuitBreakerOpen

EXPIRE = 0.02


class CustomError(Exception):
    pass


async def test_circuit_breaker_simple(cache):
    @cache.circuit_breaker(ttl=EXPIRE * 10, min_calls=10, errors_rate=5, period=1, key="test")
    async def func(fail=False):
        if fail:
            raise CustomError()
        return b"ok"

    for _ in range(9):
        assert await func() == b"ok"

    with pytest.raises(CustomError):
        await func(fail=True)
    await asyncio.sleep(0)

    with pytest.raises(CircuitBreakerOpen):
        await func(fail=True)

    with pytest.raises(CircuitBreakerOpen):
        await func(fail=False)


async def test_circuit_breaker_half_open(cache):
    @cache.circuit_breaker(
        ttl=EXPIRE,
        half_open_ttl=0.1,
        errors_rate=1,
        min_calls=0,
        period=1,
        key="test",
    )
    async def func(fail=False):
        if fail:
            raise CustomError()
        return b"ok"

    with pytest.raises(CustomError):
        await func(fail=True)

    with pytest.raises(CircuitBreakerOpen):
        await func(fail=True)

    await asyncio.sleep(EXPIRE)

    errors = 0
    success = 0
    for _ in range(100):
        try:
            await func(fail=False)
        except CircuitBreakerOpen:
            errors += 1
        else:
            success += 1
    assert success > 0
    assert errors > 0
    assert success + errors == 100

    await asyncio.sleep(0.1)
    assert await func() == b"ok"



from unittest.mock import Mock

import pytest

from cashews import Cache, only_exceptions, with_exceptions

EXPIRE = 0.1


class CustomError(Exception):
    pass


async def test_cache_with_exceptions(cache: Cache):
    mock = Mock()

    @cache(ttl=EXPIRE, condition=with_exceptions(CustomError))
    async def func(raise_exception=None):
        if raise_exception:
            mock()
            raise raise_exception
        mock()
        return b"ok"

    assert await func() == b"ok"
    assert await func() == b"ok"

    with pytest.raises(CustomError):
        await func(CustomError())

    with pytest.raises(CustomError):
        await func(CustomError())

    with pytest.raises(Exception):
        await func(Exception())

    assert mock.call_count == 3


async def test_cache_exc_not_cached_by_default(cache: Cache):
    mock = Mock()

    @cache(ttl=EXPIRE)
    async def func():
        mock()
        raise CustomError()

    with pytest.raises(CustomError):
        await func()

    with pytest.raises(CustomError):
        await func()

    assert mock.call_count == 2


async def test_cache_only_exceptions(cache: Cache):
    mock = Mock()

    @cache(ttl=EXPIRE, condition=only_exceptions(CustomError))
    async def func(raise_exception=None):
        if raise_exception:
            mock()
            raise raise_exception
        mock()
        return b"ok"

    assert await func() == b"ok"
    assert await func() == b"ok"

    with pytest.raises(CustomError):
        await func(CustomError())

    with pytest.raises(CustomError):
        await func(CustomError())

    assert mock.call_count == 3


async def test_cache_early_exceptions(cache: Cache):
    mock = Mock()

    @cache.early(ttl=EXPIRE, condition=only_exceptions(CustomError))
    async def func(raise_exception=None):
        if raise_exception:
            mock()
            raise raise_exception

    with pytest.raises(CustomError):
        await func(CustomError())

    with pytest.raises(CustomError):
        await func(CustomError())

    mock.assert_called_once()


async def test_cache_hit_exceptions(cache: Cache):
    mock = Mock()

    @cache.hit(ttl=EXPIRE, cache_hits=10, condition=only_exceptions(CustomError))
    async def func(raise_exception=None):
        if raise_exception:
            mock()
            raise raise_exception

    with pytest.raises(CustomError):
        await func(CustomError())

    with pytest.raises(CustomError):
        await func(CustomError())

    mock.assert_called_once()



from __future__ import annotations

from functools import wraps
from typing import TYPE_CHECKING, Callable

from cashews.backends.interface import _BackendInterface
from cashews.key import get_cache_key, get_cache_key_template
from cashews.ttl import ttl_to_seconds

from .defaults import _empty, context_cache_detect

if TYPE_CHECKING:  # pragma: no cover
    from cashews._typing import TTL, CallableCacheCondition, DecoratedFunc, KeyOrTemplate

__all__ = ("failover", "fast_condition")


def fast_condition(getter, setter=None):
    def _fast_condition(result, args, kwargs, key=""):
        if getter(key):
            return False
        if setter:
            setter(key, result)
        return True

    return _fast_condition


def failover(
    backend: _BackendInterface,
    ttl: TTL,
    exceptions: type[Exception] | tuple[type[Exception]] = Exception,
    key: KeyOrTemplate | None = None,
    condition: CallableCacheCondition = lambda *args, **kwargs: True,
    prefix: str = "fail",
) -> Callable[[DecoratedFunc], DecoratedFunc]:
    """
    Return cache result (at list 1 call of function call should be succeed) if call raised one of given exception,
    :param backend: cache backend
    :param ttl: duration in seconds to store a result
    :param condition: callable object that determines whether the result will be saved or not
    :param exceptions: exceptions at which returned cache result
    :param key: custom cache key, may contain alias to args or kwargs passed to a call
    :param prefix: custom prefix for key, default "fail"
    """

    ttl = ttl_to_seconds(ttl)

    def _decor(func: DecoratedFunc) -> DecoratedFunc:
        _key_template = get_cache_key_template(func, key=key, prefix=prefix)

        @wraps(func)
        async def _wrap(*args, **kwargs):
            _cache_key = get_cache_key(func, _key_template, args, kwargs)
            try:
                result = await func(*args, **kwargs)
            except exceptions as exc:
                cached = await backend.get(_cache_key, default=_empty)
                if cached is not _empty:
                    _ttl = ttl_to_seconds(ttl, *args, **kwargs, result=cached, with_callable=True)
                    context_cache_detect._set(
                        _cache_key,
                        ttl=_ttl,
                        exc=exc,
                        name="failover",
                        template=_key_template,
                        value=cached,
                    )
                    return cached
                raise exc
            else:
                if condition(result, args, kwargs, key=_cache_key):
                    _ttl = ttl_to_seconds(ttl, *args, **kwargs, result=result, with_callable=True)
                    await backend.set(_cache_key, result, expire=_ttl)
                return result

        return _wrap  # type: ignore[return-value]

    return _decor



from unittest.mock import ANY

import pytest

from cashews import Cache
from cashews.helpers import add_prefix


@pytest.fixture(autouse=True)
def _add_prefix(cache: Cache, target):
    cache.add_middleware(add_prefix("prefix!"))


async def test_add_prefix_get(cache: Cache, target):
    await cache.get(key="key")
    target.get.assert_called_once_with(key="prefix!key", default=None)


async def test_add_prefix_set(cache: Cache, target):
    await cache.set(key="key", value="value")
    target.set.assert_called_once_with(
        key="prefix!key",
        value=ANY,
        exist=None,
        expire=None,
    )


async def test_add_prefix_ping(cache: Cache, target):
    await cache.ping()
    target.ping.assert_called_once_with(message=b"PING")


async def test_add_prefix_get_many(cache: Cache, target):
    await cache.get_many("key")
    target.get_many.assert_called_once_with("prefix!key")


async def test_add_prefix_set_many(cache: Cache, target):
    await cache.set_many({"key": "value"})
    target.set_many.assert_called_once_with(pairs={"prefix!key": ANY}, expire=None)


async def test_add_prefix_delete(cache: Cache, target):
    await cache.delete("key")
    target.delete.assert_called_once_with(key="prefix!key")


async def test_add_prefix_delete_match(cache: Cache, target):
    await cache.delete_match("key")
    target.delete_match.assert_called_once_with(pattern="prefix!key")


async def test_add_prefix_delete_many(cache: Cache, target):
    await cache.delete_many("key")
    target.delete_many.assert_called_once_with("prefix!key")



from prometheus_client import Counter, Histogram

from cashews import cache
from cashews._typing import Middleware
from cashews.backends.interface import Backend
from cashews.commands import Command

_DEFAULT_METRIC = Histogram(
    "cashews_operations_latency_seconds",
    "Latency of different operations with a cache",
    labelnames=["operation", "backend_class", "tag"],
)
_HIT_MISS = Counter(
    "cashews_get_operation",
    "Count of hits or missed GET operations",
    labelnames=["result", "backend_class", "tag"],
)


def create_metrics_middleware(with_tag: bool = False) -> Middleware:
    async def metrics_middleware(call, cmd: Command, backend: Backend, *args, **kwargs):
        with _DEFAULT_METRIC.time() as metric:
            tag = ""

            if with_tag and "key" in kwargs:
                tags = cache.get_key_tags(kwargs["key"])
                tag = next((t for t in tags), "")
            metric.labels(operation=cmd.value, backend_class=backend.__class__.__name__, tag=tag)
            result = await call(*args, **kwargs)
            if cmd is Command.GET:
                op_result = "hit" if result is not kwargs["default"] else "miss"
                _HIT_MISS.labels(result=op_result, backend_class=backend.__class__.__name__, tag=tag).inc()
            return result

    return metrics_middleware



from unittest.mock import Mock

import pytest

from cashews.decorators.bloom import bloom


@pytest.fixture
async def backend(raw_backend):
    _backend, _ = raw_backend
    yield _backend


async def test_bloom_simple(cache):
    n = 100
    call = Mock()

    @cache.bloom(name="name:{k}", false_positives=1, capacity=n)
    async def func(k):
        call(k)
        return k > (n / 2)

    for i in range(n):
        await func.set(i)

    for i in range(n):
        call.reset_mock()
        assert await func(i) is (i > (n / 2))
        if i > (n / 2):
            call.assert_called_with(i)
        else:
            call.assert_not_called()


async def test_bloom_simple_big_size(backend):
    n = 1_000_000
    call = Mock()

    @bloom(backend=backend, name="name:{k}", false_positives=1, capacity=n)
    async def func(k):
        call(k)
        return k > (n / 2)

    await func.set(900_000)
    await func.set(1_000)

    call.reset_mock()
    assert await func(900_000)
    call.assert_called_with(900_000)

    call.reset_mock()
    assert not await func(1_000)
    call.assert_not_called()


async def test_bloom_dual(cache):
    n = 100
    call = Mock()

    @cache.dual_bloom(name="name:{k}", false=1, capacity=n)
    async def func(k):
        call(k)
        return k > (n / 2)

    for i in range(n - 1):
        await func(i)

    call.reset_mock()
    for i in range(n - 1):
        assert await func(i) is (i > (n / 2))

    assert call.call_count == 0

    assert await func(100) is True
    call.assert_called_with(100)



import dataclasses
from collections import namedtuple
from datetime import date, datetime, timedelta
from decimal import Decimal

import pytest
from hypothesis import example, given, settings
from hypothesis import strategies as st

from cashews.backends.memory import Memory
from cashews.picklers import PicklerType
from cashews.serialize import UnSecureDataError, get_serializer


@dataclasses.dataclass()
class TestDC:
    test: str
    _: int


NT = namedtuple("NT", ("test", "name"), defaults=[False, "test"])


@pytest.fixture(
    name="cache",
    params=[
        "default_md5",
        "default_sum",
        "default_sha256",
        pytest.param("redis_md5", marks=pytest.mark.redis),
        pytest.param("redis_sum", marks=pytest.mark.redis),
        pytest.param("dill_sum", marks=pytest.mark.integration),
        pytest.param("sqlalchemy_sha1", marks=pytest.mark.integration),
    ],
)
async def _cache(request, redis_dsn):
    pickle_type, digestmod = request.param.split("_")
    if pickle_type == "redis":
        from cashews.backends.redis import Redis

        redis = Redis(redis_dsn, suppress=False, serializer=get_serializer(secret=b"test", digestmod=digestmod))
        await redis.init()
        await redis.clear()
        yield redis
        await redis.close()
    else:
        yield Memory(serializer=get_serializer(secret=b"test", digestmod=digestmod))


@pytest.mark.parametrize(
    "value",
    (
        "test",
        b"test",
        "1_1",
        b"1_1",
        b"1_1_1",
        "1",
        b"0",
        0,
        1,
        2,
        1.234,
        Decimal("1.001"),
        True,
        False,
        "",
        None,
        {"hi": True},
        TestDC(test="test", _=1),
    ),
)
async def test_serialize_simple_value(value, cache):
    await cache.set("key", value)
    await cache.set_many({"key1": value, "key2": value})
    assert await cache.get("key") == value
    assert await cache.get("key2") == value
    assert await cache.get_many("key1", "key2") == (value, value)


@pytest.mark.parametrize("value", (NT(name="test", test="lol"), NT()))
async def test_serialize_collections_value(value, cache):
    await cache.set("key", value)
    assert await cache.get("key") == value
    assert (await cache.get("key")).name == "test"


@pytest.mark.parametrize(
    "value",
    (
        datetime(year=2000, month=1, day=10, hour=10),
        timedelta(days=10),
        date(year=2020, month=12, day=31),
    ),
)
async def test_serialize_dates_value(value, cache):
    await cache.set("key", value)
    assert await cache.get("key") == value


@pytest.mark.parametrize(
    "value",
    (
        ["test", "to"],
        (TestDC("hay", _=2),),
        {"test"},
        [(1, 2), (3, 4)],
        {1, 2, 5},
        [{"test": True}],
    ),
)
async def test_serialize_array_value(value, cache):
    await cache.set("key", value)
    assert list(await cache.get("key")) == list(value)
    assert type(await cache.get("key")) is type(value)


@pytest.mark.parametrize(
    "value",
    (
        ["test", b"to"],
        (TestDC("hay", _=2), Decimal("10.1")),
        {"test", 1},
        [],
        [{1: True}, Decimal("0.1")],
    ),
)
async def test_serialize_array_diff_value(value, cache):
    await cache.set("key", value)
    assert list(await cache.get("key")) == list(value)
    assert type(await cache.get("key")) is type(value)


@pytest.mark.parametrize(
    ("value", "encoded"),
    (
        (
            "test",
            b"md5:2182de23a749c70144594da72f8cedc2_\x80\x05\x95\b\x00\x00\x00\x00\x00\x00\x00\x8c\x04test\x94.",
        ),
        (b"test", b"md5:aa47dd7b9ec8a973cdea43bce56cd34c_bytes:test"),
        (0, b"0"),
        (
            1.234,
            b"sum:915_\x80\x05\x95\n\x00\x00\x00\x00\x00\x00\x00G?\xf3\xbev\xc8\xb49X.",
        ),
        (
            Decimal("1.001"),
            b"sum:117b_\x80\x05\x95#\x00\x00\x00\x00\x00\x00\x00\x8c\adecimal\x94\x8c\aDecimal\x94\x93\x94\x8c\x05"
            b"1.001\x94\x85\x94R\x94.",
        ),
        (True, b"md5:b5d04289f61dd9a7cb8856222e64589a_\x80\x05\x88."),
        (None, b"sum:40a_\x80\x05N."),
        ("1", b"sum:5a8_\x80\x05\x95\x05\x00\x00\x00\x00\x00\x00\x00\x8c\x011\x94."),
        (
            {"hi": True},
            b"sum:85a_\x80\x05\x95\n\x00\x00\x00\x00\x00\x00\x00}\x94\x8c\x02hi\x94\x88s.",
        ),
        (
            TestDC(test="test", _=1),
            b"sum:1e3f_\x80\x05\x95B\x00\x00\x00\x00\x00\x00\x00\x8c\x1ctests.test_pickle_serializer\x94\x8c\x06"
            b"TestDC\x94\x93\x94)\x81\x94}\x94(\x8c\x04test\x94h\x05\x8c\x01_\x94K\x01ub.",
        ),
        (
            [
                NT(name="test", test="lol"),
                1,
                timedelta(days=10),
                TestDC(test="test", _=1),
            ],
            b"sum:360b_\x80\x05\x95\x80\x00\x00\x00\x00\x00\x00\x00]\x94(\x8c\x1ctests.test_pickle_serializer"
            b"\x94\x8c\x02NT\x94\x93\x94\x8c\x03lol\x94\x8c\x04test\x94\x86\x94\x81\x94K\x01\x8c\bdatetime"
            b"\x94\x8c\ttimedelta\x94\x93\x94K\nK\x00K\x00\x87\x94R\x94h\x01\x8c\x06TestDC"
            b"\x94\x93\x94)\x81\x94}\x94(h\x05h\x05\x8c\x01_\x94K\x01ube.",
        ),
        (b"1", b"md5:fa81eb7dc6d56848cb9b3dab70437927_bytes:1"),
        (1, b"1"),
        ("", b"sum:575_\x80\x05\x95\x04\x00\x00\x00\x00\x00\x00\x00\x8c\x00\x94."),
    ),
)
async def test_serialize_backward(value, encoded, cache):
    await cache.set_raw("key", encoded)
    assert await cache.get("key") == value


@pytest.mark.parametrize(
    "value",
    (
        b"_cos\nsystem\n(S'echo hello world'\ntR.",
        b":_cos\nsystem\n(S'echo hello world'\ntR.",
        b"md5:_cos\nsystem\n(S'echo hello world'\ntR.",
        b"sha1:_cos\nsystem\n(S'echo hello world'\ntR.",
        b"__cos\nsystem\n(S'echo hello world'\ntR.",
    ),
)
async def test_unsecure_value(value, cache):
    await cache.set_raw("key", value)
    with pytest.raises(UnSecureDataError):
        assert not await cache.get("key")


async def test_unsecure_value_many(cache):
    await cache.set_raw("key", b"_cos\nsystem\n(S'echo hello world'\ntR.")
    with pytest.raises(UnSecureDataError):
        await cache.get_many("key")


async def test_no_value(cache):
    assert await cache.get("key", default="default") == "default"


async def test_replace_values(cache):
    await cache.set("key", "key")
    await cache.set_raw("replace", await cache.get_raw("key"))

    with pytest.raises(UnSecureDataError):
        await cache.get("replace")


async def test_pickle_error_value(cache):
    await cache.set_raw("key", b"nopickledata")
    assert await cache.get("key", default="default") == "default"


async def test_set_no_ser(cache):
    empty = object()
    await cache.set("key_e", empty)


Schema = None


async def test_data_change(cache):
    global Schema

    @dataclasses.dataclass
    class Schema:
        test: str

    await cache.set("key", Schema("test"))

    @dataclasses.dataclass
    class Schema:
        name: str

    # Check without data change
    await cache.set("key2", Schema("test"))
    value = await cache.get("key2")
    assert value.name == "test"

    # IN key class of data have changed so it should be invalid
    value = await cache.get("key")
    assert value is None


async def test_get_set_raw(cache):
    await cache.set_raw("key", b"test")
    assert await cache.get_raw("key") == b"test"


@given(key=st.text(), value=st.characters())
@settings(max_examples=500)
@example(key="_key:_!@#$%^&*()", value='_value:_!@_#$%^&:*(?)".,4й')
async def test_no_hash(key, value):
    cache = Memory(serializer=get_serializer())
    await cache.set(key, value)
    assert await cache.get(key) == value


async def test_cache_from_hash_to_no_hash():
    val = Decimal("10.2")
    cache = Memory(serializer=get_serializer(secret="test"))
    await cache.set("key", val)

    assert await cache.get("key") == val
    cache_no_hash = Memory(serializer=get_serializer())
    cache_no_hash.store = cache.store
    assert await cache_no_hash.get("key", default="default") == "default"


async def test_cache_from_no_hash_to_hash():
    val = Decimal("10.2")
    cache = Memory(serializer=get_serializer())
    await cache.set("key", val)

    assert await cache.get("key") == val
    cache_hash = Memory(serializer=get_serializer(secret="test"))
    cache_hash.store = cache.store

    assert await cache_hash.get("key") == val


@pytest.mark.parametrize(
    "value",
    (
        "str",
        123,
        [1, 2, 3],
        ["s", "t", "r"],
        {"any": None},
    ),
)
async def test_json_serialize(value):
    cache = Memory(serializer=get_serializer(secret=b"test", pickle_type=PicklerType.JSON))
    await cache.set("key", value)
    assert await cache.get("key") == value



from __future__ import annotations

import contextlib
from collections.abc import Sequence
from contextlib import AbstractContextManager, nullcontext
from contextvars import ContextVar
from datetime import datetime, timezone
from hashlib import blake2s
from typing import Any

from starlette.middleware.base import BaseHTTPMiddleware, RequestResponseEndpoint
from starlette.requests import Request
from starlette.responses import Response
from starlette.types import ASGIApp

from cashews import Cache, Command, cache, invalidate_further
from cashews._typing import TTL
from cashews.picklers import DEFAULT_PICKLER
from cashews.ttl import ttl_to_seconds

_cache_max_age: ContextVar[int] = ContextVar("cache_control_max_age")

PREFIX = "fastapi"
_CACHE_CONTROL_HEADER = "Cache-Control"
_AGE_HEADER = "Age"
_ETAG_HEADER = "ETag"
_IF_NOT_MATCH_HEADER = "If-None-Match"
_CLEAR_CACHE_HEADER = "Clear-Site-Data"

_NO_CACHE = "no-cache"  # disable GET
_NO_STORE = "no-store"  # disable GET AND SET
_MAX_AGE = "max-age="
_PRIVATE = "private"
_PUBLIC = "public"

_CLEAR_CACHE_HEADER_VALUE = "cache"
__all__ = [
    "cache_control_ttl",
    "CacheRequestControlMiddleware",
    "CacheEtagMiddleware",
    "CacheDeleteMiddleware",
]


def cache_control_ttl(default: TTL):
    def _ttl(*args, **kwargs):
        return _cache_max_age.get(ttl_to_seconds(default))

    return _ttl


class CacheRequestControlMiddleware(BaseHTTPMiddleware):
    def __init__(
        self,
        app: ASGIApp,
        cache_instance: Cache = cache,
        methods: Sequence[str] = ("get",),
        private=True,
        prefix_to_disable: str = "",
    ):
        self._private = private
        self._cache = cache_instance
        self._methods = methods
        self._prefix_to_disable = prefix_to_disable
        super().__init__(app)

    async def dispatch(self, request: Request, call_next: RequestResponseEndpoint) -> Response:
        context: AbstractContextManager = nullcontext()
        cache_control_value = request.headers.get(_CACHE_CONTROL_HEADER)
        if request.method.lower() not in self._methods:
            return await call_next(request)
        to_disable = _to_disable(cache_control_value)
        if to_disable:
            context = self._cache.disabling(*to_disable, prefix=self._prefix_to_disable)
        with context, max_age(cache_control_value), self._cache.detect as detector:
            response = await call_next(request)
            calls = detector.calls_list
            if calls:
                key, _ = calls[0]
                expire = await self._cache.get_expire(key)
                if expire > 0:
                    response.headers[_CACHE_CONTROL_HEADER] = (
                        f"{_PRIVATE if self._private else _PUBLIC}, {_MAX_AGE}{expire}"
                    )
                    response.headers[_AGE_HEADER] = f"{expire}"
            else:
                response.headers[_CACHE_CONTROL_HEADER] = _NO_CACHE
            return response


@contextlib.contextmanager
def max_age(cache_control_value: str | None):
    if not cache_control_value:
        yield
        return
    _max_age = _get_max_age(cache_control_value)
    reset_token = None
    if _max_age:
        reset_token = _cache_max_age.set(_max_age)
    try:
        yield
    finally:
        if reset_token:
            _cache_max_age.reset(reset_token)


def _to_disable(cache_control_value: str | None) -> tuple[Command, ...]:
    if cache_control_value == _NO_CACHE:
        return (Command.GET,)
    if cache_control_value == _NO_STORE:
        return Command.GET, Command.SET
    if cache_control_value and _get_max_age(cache_control_value) == 0:
        return Command.GET, Command.SET
    return ()


def _get_max_age(cache_control_value: str) -> int | None:
    if _MAX_AGE not in cache_control_value:
        return None
    for cc_value_item in cache_control_value.split(","):
        cc_value_item = cc_value_item.strip()
        try:
            key, value = cc_value_item.split("=", maxsplit=1)
            if key == _MAX_AGE[:-1]:
                return int(value)
        except (ValueError, TypeError):
            continue
    return None


class CacheEtagMiddleware(BaseHTTPMiddleware):
    def __init__(self, app: ASGIApp, cache_instance: Cache = cache, prefix: str = PREFIX):
        self._cache = cache_instance
        self._prefix = prefix
        super().__init__(app)

    async def dispatch(self, request: Request, call_next: RequestResponseEndpoint) -> Response:
        request_etag = request.headers.get(_IF_NOT_MATCH_HEADER)
        request_etag = request_etag.replace('"', "") if request_etag else None
        if request_etag and await self._cache.exists(self._get_etag_key(request_etag)):
            return Response(status_code=304)

        used_key: None | str = None

        def set_callback(key: str, result: Any):
            if not key.endswith(":lock"):
                nonlocal used_key
                used_key = key

        with self._cache.detect as detector, self._cache.callback(set_callback, cmd=Command.SET):
            response = await call_next(request)
            calls = detector.calls_list
            if not calls and not used_key:
                return response
            if used_key:
                _key = used_key
                _data = None
            else:
                _key = calls[0][0]
                _data = calls[0][1]["value"]
            _etag = await self._set_etag(_key, _data)
            return self._response_etag(response, _etag, request_etag)

    def _response_etag(self, response: Response, etag: str | None, request_etag: str | None) -> Response:
        if etag is None:
            return response
        if etag == request_etag:
            return Response(status_code=304)
        if etag:
            response.headers[_ETAG_HEADER] = f'"{etag}"'
        return response

    async def _set_etag(self, key: str, data: Any = None) -> str | None:
        _data = data or await self._cache.get(key)
        if _is_early_cache(_data):
            expire = (_data[0] - datetime.now(timezone.utc)).total_seconds()  # type: ignore[index]
            _data = _data[1]  # type: ignore[index]
        else:
            expire = await self._cache.get_expire(key)
        if expire < 0:
            return None
        etag = _get_etag(_data)
        await self._cache.set(self._get_etag_key(etag), True, expire=expire)
        return etag

    def _get_etag_key(self, etag: str) -> str:
        return f"{self._prefix}:etag:{etag}"


def _get_etag(cached_data: Any) -> str:
    if not isinstance(cached_data, bytes):
        cached_data = cached_data.body if isinstance(cached_data, Response) else DEFAULT_PICKLER.dumps(cached_data)
    return blake2s(cached_data).hexdigest()


def _is_early_cache(data: Any) -> bool:
    return bool(isinstance(data, list) and data and isinstance(data[0], datetime))


class CacheDeleteMiddleware(BaseHTTPMiddleware):
    async def dispatch(self, request: Request, call_next: RequestResponseEndpoint):
        if request.headers.get(_CLEAR_CACHE_HEADER) == _CLEAR_CACHE_HEADER_VALUE:
            with invalidate_further():
                return await call_next(request)
        return await call_next(request)



from __future__ import annotations

import base64
import json
import re
from collections.abc import Iterable
from contextlib import contextmanager
from hashlib import md5, sha1, sha256
from re import Pattern
from string import Formatter
from typing import Any, Callable

from . import key_context
from ._typing import KeyOrTemplate, KeyTemplate

TemplateValue = str
_re_special_chars_map = {i: "\\" + chr(i) for i in b"()[]?*+-|^$\\&~# \t\n\r\v\f"}


def _decode_bytes(value: bytes):
    try:
        return value.decode()
    except UnicodeDecodeError:
        return value.hex()


def _decode_exception(value: Exception):
    return f"{value.__class__.__name__}:{value}"


def _get_decode_array(format_value):
    def _decode_array(values: Iterable[str]) -> TemplateValue:
        return ":".join([format_value(value) for value in values])

    return _decode_array


def _get_decoded_dict(format_value):
    def _decode_dict(value: dict) -> TemplateValue:
        _kv = (format_value(k) + ":" + format_value(v) for k, v in sorted(value.items()))
        return ":".join(_kv)

    return _decode_dict


def _decoded_bool(value: bool) -> TemplateValue:
    return str(value).lower()


def _decode_direct(value: str) -> TemplateValue:
    return value


class _ReplaceFormatter(Formatter):
    def __init__(self, default=lambda field: "*"):
        self.__default = default
        _decode_array = _get_decode_array(self._format_field)
        self.__type_format = {
            str: _decode_direct,
            bool: _decoded_bool,
            bytes: _decode_bytes,
            Exception: _decode_exception,
            tuple: _decode_array,
            list: _decode_array,
            set: _decode_array,
            dict: _get_decoded_dict(self._format_field),
        }
        super().__init__()

    @contextmanager
    def default(self, _default):
        was = self.__default
        self.__default = _default
        try:
            yield
        finally:
            self.__default = was

    def set_format_for_type(self, value, format_function):
        self.__type_format[value] = format_function

    def type_format(self, value):
        def _decorator(func):
            self.set_format_for_type(value, func)
            return func

        return _decorator

    def get_field(self, field_name, args, kwargs):
        try:
            return super().get_field(field_name, args, kwargs)
        except (KeyError, AttributeError):
            return self.__default(field_name), None

    def _format_field(self, value: Any):
        if value is None:
            return ""
        return str(self._type_format(value))

    def _type_format(self, value: Any):
        _type = type(value)
        if _type in self.__type_format:
            return self.__type_format[_type](value)
        for _type_map, func_format in self.__type_format.items():
            if isinstance(value, _type_map):
                self.__type_format[_type] = func_format  # to avoid isinstance next time
                return func_format(value)
        return str(value)

    def format_field(self, value, format_spec):
        return format(self._format_field(value))


class _FuncFormatter(_ReplaceFormatter):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        self._functions: dict[str, tuple[Callable, bool]] = {}
        super().__init__(*args, **kwargs)

    def _register(self, alias: str, function: Callable, preformat: bool = True) -> None:
        self._functions[alias] = (function, preformat)

    def register(self, alias: str, preformat: bool = True):
        def _decorator(func):
            self._register(alias, func, preformat=preformat)
            return func

        return _decorator

    def format_field(self, value: Any, format_spec: str) -> TemplateValue:
        if format_spec == "":
            return super().format_field(value, format_spec)
        format_spec, args = self.parse_format_spec(format_spec)
        if format_spec not in self._functions:
            return super().format_field(value, format_spec)
        func, preformat = self._functions[format_spec]
        if preformat:
            value = super().format_field(value, "")
        return func(value, *args)

    @staticmethod
    def parse_format_spec(format_spec: str):
        if not format_spec or "(" not in format_spec:
            return format_spec, ()
        format_spec, args = format_spec.split("(", maxsplit=1)
        return format_spec, args.replace(")", "").split(",")

    def vformat(self, format_string, args, kwargs):
        try:
            return format_string.format(**{key: self._type_format(val) for key, val in kwargs.items()})
        except (ValueError, TypeError, KeyError, AttributeError):
            return super().vformat(format_string, args, kwargs)


default_formatter = _FuncFormatter(lambda name: "")


@default_formatter.register("get", preformat=False)
def _get(value: Any, key: str) -> TemplateValue:
    return str(value.get(key))


@default_formatter.register("len")
def _len(value: TemplateValue):
    return str(len(value))


@default_formatter.register("jwt")
def _jwt_func(jwt: TemplateValue, key: str) -> TemplateValue:
    _, payload, _ = jwt.split(".", maxsplit=2)
    payload_dict = json.loads(base64.b64decode(payload))
    return payload_dict.get(key)


@default_formatter.register("hash")
def _hash_func(value: TemplateValue, alg="md5") -> TemplateValue:
    algs = {"sha1": sha1, "md5": md5, "sha256": sha256}
    alg = algs[alg]
    return alg(value.encode()).hexdigest()


@default_formatter.register("lower")
def _lower(value: TemplateValue) -> TemplateValue:
    return value.lower()


@default_formatter.register("upper")
def _upper(value: TemplateValue) -> TemplateValue:
    return value.upper()


def default_format(template: KeyTemplate, **values) -> KeyOrTemplate:
    _template_context, rewrite = key_context.get()
    if _template_context:
        values["@"] = _template_context
        if rewrite:
            _template_context = {**values, **_template_context}
        else:
            _template_context = {**_template_context, **values}
    else:
        _template_context = values
    return default_formatter.format(template, **_template_context)


def _re_default(field_name):
    field_name = field_name.split(".")[0]
    return f"(?P<{field_name}>.+)?"


_re_formatter = _ReplaceFormatter(default=_re_default)


def template_to_re_pattern(template: KeyTemplate) -> Pattern:
    pattern = _re_formatter.format(template.translate(_re_special_chars_map))
    return re.compile("^" + pattern + "$", flags=re.MULTILINE)



from __future__ import annotations

import asyncio
import logging
from datetime import datetime, timedelta, timezone
from functools import wraps
from typing import TYPE_CHECKING, Callable

from cashews.key import get_cache_key, get_cache_key_template
from cashews.ttl import ttl_to_seconds

from ._exception import RaiseException, return_or_raise
from .defaults import _empty, context_cache_detect

if TYPE_CHECKING:  # pragma: no cover
    from cashews import Cache
    from cashews._typing import TTL, CallableCacheCondition, DecoratedFunc, KeyOrTemplate, Tags

__all__ = ("early",)

logger = logging.getLogger(__name__)
_LOCK_SUFFIX = ":lock"


def early(
    backend: Cache,
    ttl: TTL,
    key: KeyOrTemplate | None = None,
    early_ttl: TTL | None = None,
    condition: CallableCacheCondition = lambda *args, **kwargs: True,
    prefix: str = "early",
    tags: Tags = (),
    background: bool = True,
) -> Callable[[DecoratedFunc], DecoratedFunc]:
    """
    Cache strategy that try to solve Cache stampede problem (https://en.wikipedia.org/wiki/Cache_stampede),
    With hot cache recalculate a result near expiration time
    Warning Not good at cold cache

    :param backend: cache backend
    :param ttl: duration in seconds to store a result
    :param key: custom cache key, may contain alias to args or kwargs passed to a call
    :param early_ttl: duration in seconds to expire results
    :param condition: callable object that determines whether the result will be saved or not
    :param prefix: custom prefix for key, default 'early'
    :param tags: aliases for keys that used for cache (used for invalidation)
    :param background: if true will run recalculation in background
    """

    background_tasks = set()
    ttl = ttl_to_seconds(ttl)
    early_ttl = ttl_to_seconds(early_ttl)

    def _decor(func: DecoratedFunc) -> DecoratedFunc:
        _key_template = get_cache_key_template(func, key=key, prefix=prefix + ":v2")
        for tag in tags:
            backend.register_tag(tag, _key_template)

        @wraps(func)
        async def _wrap(*args, **kwargs):
            _ttl = ttl_to_seconds(ttl, *args, **kwargs, with_callable=True)
            if early_ttl is None:
                _early_ttl = _ttl * 0.33
            else:
                _early_ttl = ttl_to_seconds(early_ttl, *args, **kwargs, with_callable=True)
            _cache_key = get_cache_key(func, _key_template, args, kwargs)
            _tags = [get_cache_key(func, tag, args, kwargs) for tag in tags]

            args_to_call = [
                backend,
                func,
                args,
                kwargs,
                _cache_key,
                _ttl,
                _early_ttl,
                condition,
                _tags,
            ]
            cached = await backend.get(_cache_key, default=_empty)
            if cached is _empty:
                return await _get_result_for_early(*args_to_call)

            early_expire_at, result = cached
            context_cache_detect._set(
                _cache_key,
                ttl=_ttl,
                early_ttl=_early_ttl,
                name="early",
                template=_key_template,
                value=result,
                early_expire_at=early_expire_at,
            )
            if early_expire_at >= datetime.now(timezone.utc):
                return return_or_raise(result)
            lock_key = _cache_key + _LOCK_SUFFIX
            if not await backend.set(lock_key, "1", expire=_early_ttl, exist=False):
                return return_or_raise(result)
            logger.info(
                "Recalculate cache for %s (exp_at %s)",
                _cache_key,
                early_expire_at,
            )
            task = asyncio.create_task(_get_result_for_early(*args_to_call, unlock=True))
            background_tasks.add(task)
            task.add_done_callback(background_tasks.discard)
            if not background:
                await task
            return return_or_raise(result)

        return _wrap  # type: ignore[return-value]

    return _decor


async def _get_result_for_early(
    backend: Cache,
    func,
    args,
    kwargs,
    key,
    ttl: int,
    early_ttl: int,
    condition,
    tags,
    unlock=False,
):
    try:
        _exc = None
        try:
            result = await func(*args, **kwargs)
        except Exception as exc:
            _exc = exc
            result = exc
        cond_result = condition(result, args, kwargs, key=key)
        early_expire_at = datetime.now(timezone.utc) + timedelta(seconds=early_ttl)
        if isinstance(cond_result, bool) and cond_result and not isinstance(result, Exception):
            await backend.set(key, [early_expire_at, result], expire=ttl, tags=tags)
        elif isinstance(cond_result, Exception):
            await backend.set(key, [early_expire_at, RaiseException(result)], expire=ttl, tags=tags)
        if _exc:
            raise _exc
        return return_or_raise(result)
    finally:
        if unlock:
            asyncio.create_task(backend.delete(key + _LOCK_SUFFIX))



from __future__ import annotations

from functools import wraps
from typing import TYPE_CHECKING, Callable

from cashews import decorators, validation
from cashews.cache_condition import get_cache_condition
from cashews.ttl import ttl_to_seconds

from .time_condition import create_time_condition
from .wrapper import Wrapper

if TYPE_CHECKING:  # pragma: no cover
    from cashews._typing import TTL, AsyncCallable_T, CacheCondition, DecoratedFunc, Exceptions, KeyOrTemplate, Tags
    from cashews.decorators.bloom import IntOrPair


def _skip_thunder_protection(func: DecoratedFunc) -> DecoratedFunc:
    return func


class DecoratorsWrapper(Wrapper):
    _default_fail_exceptions: tuple[type[Exception], ...] = (Exception,)

    def set_default_fail_exceptions(self, *exc: type[Exception]) -> None:
        self._default_fail_exceptions = exc

    def _wrap_on(self, decorator_fabric, upper: bool, protected=False, **decor_kwargs):
        if decor_kwargs.get("lock") and "ttl" in decor_kwargs and decor_kwargs["ttl"] is None:
            raise ValueError("ttl can't be None with lock")

        if upper:
            return self._wrap_with_condition(decorator_fabric, **decor_kwargs)
        return self._wrap(decorator_fabric, protected=protected, **decor_kwargs)

    def _wrap(
        self,
        decorator_fabric,
        lock=False,
        time_condition=None,
        protected=False,
        **decor_kwargs,
    ) -> Callable[[DecoratedFunc], DecoratedFunc]:
        def _decorator(func: DecoratedFunc) -> DecoratedFunc:
            if time_condition is not None:
                condition, _decor = create_time_condition(time_condition)
                func = _decor(func)
                decor_kwargs["condition"] = condition

            decorator = decorator_fabric(self, **decor_kwargs)(func)
            thunder_protection: Callable[[DecoratedFunc], DecoratedFunc] = _skip_thunder_protection
            if protected:
                thunder_protection = decorators.thunder_protection(key=decor_kwargs.get("key"))

            @wraps(func)
            async def _call(*args, **kwargs):
                self._check_setup()
                if self.is_full_disable:
                    return await func(*args, **kwargs)
                if lock:
                    _locked = decorators.locked(
                        backend=self,
                        key=decor_kwargs.get("key"),
                        ttl=decor_kwargs["ttl"],
                        wait=True,
                    )
                    return await thunder_protection(_locked(decorator))(*args, **kwargs)
                else:
                    return await thunder_protection(decorator)(*args, **kwargs)

            return _call  # type: ignore[return-value]

        return _decorator

    def _wrap_with_condition(
        self,
        decorator_fabric,
        condition,
        lock=False,
        time_condition=None,
        **decor_kwargs,
    ):
        def _decorator(func: AsyncCallable_T) -> AsyncCallable_T:
            _condition = condition
            if time_condition is not None:
                _condition, _decor = create_time_condition(time_condition)
                func = _decor(func)
            decorator_fabric(self, **decor_kwargs)(func)  # to register cache templates

            @wraps(func)
            async def _call(*args, **kwargs):
                self._check_setup()
                if self.is_full_disable:
                    return await func(*args, **kwargs)
                with decorators.context_cache_detect as detect:

                    def new_condition(result, _args, _kwargs, key):
                        if detect.calls:
                            return False
                        return _condition(result, _args, _kwargs, key=key) if _condition else result is not None

                    decorator = decorator_fabric(self, **decor_kwargs, condition=new_condition)
                    if lock:
                        _locked = decorators.locked(
                            backend=self,
                            key=decor_kwargs.get("key"),
                            ttl=decor_kwargs["ttl"],
                            wait=True,
                        )
                        _result = await _locked(decorator(func))(*args, **kwargs)
                    else:
                        _result = await decorator(func)(*args, **kwargs)

                return _result

            return _call

        return _decorator

    def __call__(
        self,
        ttl: TTL,
        key: KeyOrTemplate | None = None,
        condition: CacheCondition = None,
        time_condition: TTL | None = None,
        prefix: str = "",
        upper: bool = False,
        lock: bool = False,
        tags: Tags = (),
        protected: bool = True,
    ) -> Callable[[DecoratedFunc], DecoratedFunc]:
        return self._wrap_on(
            decorators.cache,
            upper,
            lock=lock,
            ttl=ttl,
            key=key,
            condition=get_cache_condition(condition),
            time_condition=ttl_to_seconds(time_condition),
            prefix=prefix,
            tags=tags,
            protected=protected,
        )

    cache = __call__

    def failover(
        self,
        ttl: TTL,
        exceptions: Exceptions = None,
        key: KeyOrTemplate | None = None,
        condition: CacheCondition = None,
        time_condition: TTL | None = None,
        prefix: str = "fail",
    ) -> Callable[[DecoratedFunc], DecoratedFunc]:
        exceptions = exceptions or self._default_fail_exceptions
        return self._wrap_with_condition(
            decorators.failover,
            ttl=ttl,
            exceptions=exceptions,
            key=key,
            condition=get_cache_condition(condition),
            time_condition=ttl_to_seconds(time_condition),
            prefix=prefix,
        )

    def early(
        self,
        ttl: TTL,
        key: KeyOrTemplate | None = None,
        early_ttl: TTL | None = None,
        condition: CacheCondition = None,
        time_condition: TTL | None = None,
        prefix: str = "early",
        upper: bool = False,
        tags: Tags = (),
        background: bool = True,
        protected: bool = True,
    ) -> Callable[[DecoratedFunc], DecoratedFunc]:
        return self._wrap_on(
            decorators.early,
            upper,
            ttl=ttl,
            key=key,
            early_ttl=early_ttl,
            condition=get_cache_condition(condition),
            time_condition=ttl_to_seconds(time_condition),
            prefix=prefix,
            tags=tags,
            background=background,
            protected=protected,
        )

    def soft(
        self,
        ttl: TTL,
        key: KeyOrTemplate | None = None,
        soft_ttl: TTL | None = None,
        exceptions: Exceptions = Exception,
        condition: CacheCondition = None,
        time_condition: TTL | None = None,
        prefix: str = "soft",
        upper: bool = False,
        tags: Tags = (),
        protected: bool = True,
    ) -> Callable[[DecoratedFunc], DecoratedFunc]:
        return self._wrap_on(
            decorators.soft,
            upper,
            ttl=ttl,
            key=key,
            soft_ttl=ttl_to_seconds(soft_ttl),
            exceptions=exceptions,
            condition=get_cache_condition(condition),
            time_condition=ttl_to_seconds(time_condition),
            prefix=prefix,
            tags=tags,
            protected=protected,
        )

    def hit(
        self,
        ttl: TTL,
        cache_hits: int,
        update_after: int = 0,
        key: KeyOrTemplate | None = None,
        condition: CacheCondition = None,
        time_condition: TTL | None = None,
        prefix: str = "hit",
        upper: bool = False,
        tags: Tags = (),
        background: bool = True,
    ) -> Callable[[DecoratedFunc], DecoratedFunc]:
        return self._wrap_on(
            decorators.hit,
            upper,
            ttl=ttl,
            cache_hits=cache_hits,
            update_after=update_after,
            key=key,
            condition=get_cache_condition(condition),
            time_condition=ttl_to_seconds(time_condition),
            prefix=prefix,
            tags=tags,
            background=background,
        )

    def dynamic(
        self,
        ttl: TTL = 60 * 60 * 24,
        key: KeyOrTemplate | None = None,
        condition: CacheCondition = None,
        time_condition: TTL | None = None,
        prefix: str = "dynamic",
        upper: bool = False,
        tags: Tags = (),
    ) -> Callable[[DecoratedFunc], DecoratedFunc]:
        return self._wrap_on(
            decorators.hit,
            upper,
            ttl=ttl,
            cache_hits=3,
            update_after=1,
            key=key,
            condition=get_cache_condition(condition),
            time_condition=ttl_to_seconds(time_condition),
            prefix=prefix,
            tags=tags,
        )

    def iterator(
        self,
        ttl: TTL,
        key: KeyOrTemplate | None = None,
        condition: CacheCondition = None,
    ) -> Callable[[DecoratedFunc], DecoratedFunc]:
        return decorators.iterator(
            self,  # type: ignore[arg-type]
            ttl=ttl,
            key=key,
            condition=get_cache_condition(condition),
        )

    def invalidate(
        self,
        key_template: KeyOrTemplate,
        args_map: dict[str, str] | None = None,
        defaults: dict | None = None,
    ) -> Callable[[DecoratedFunc], DecoratedFunc]:
        return validation.invalidate(
            backend=self,  # type: ignore[arg-type]
            key_template=key_template,
            args_map=args_map,
            defaults=defaults,
        )

    def circuit_breaker(
        self,
        errors_rate: int,
        period: TTL,
        ttl: TTL,
        half_open_ttl: TTL = None,
        exceptions: Exceptions = None,
        key: KeyOrTemplate | None = None,
        min_calls: int = 1,
        prefix: str = "circuit_breaker",
    ) -> Callable[[DecoratedFunc], DecoratedFunc]:
        _exceptions = exceptions or self._default_fail_exceptions
        return decorators.circuit_breaker(
            backend=self,  # type: ignore[arg-type]
            errors_rate=errors_rate,
            period=ttl_to_seconds(period),
            ttl=ttl_to_seconds(ttl),
            half_open_ttl=ttl_to_seconds(half_open_ttl),
            exceptions=_exceptions,
            min_calls=min_calls,
            key=key,
            prefix=prefix,
        )

    def rate_limit(
        self,
        limit: int,
        period: TTL,
        ttl: TTL | None = None,
        action: Callable | None = None,
        prefix="rate_limit",
        key: KeyOrTemplate | None = None,
    ) -> Callable[[DecoratedFunc], DecoratedFunc]:  # pylint: disable=too-many-arguments
        return decorators.rate_limit(
            backend=self,  # type: ignore[arg-type]
            limit=limit,
            period=period,
            ttl=ttl,
            action=action,
            key=key,
            prefix=prefix,
        )

    def slice_rate_limit(
        self,
        limit: int,
        period: TTL,
        key: KeyOrTemplate | None = None,
        action: Callable | None = None,
        prefix="srl",
    ) -> Callable[[DecoratedFunc], DecoratedFunc]:
        return decorators.slice_rate_limit(
            backend=self,  # type: ignore[arg-type]
            limit=limit,
            period=period,
            key=key,
            action=action,
            prefix=prefix,
        )

    def locked(
        self,
        ttl: TTL | None = None,
        key: KeyOrTemplate | None = None,
        wait: bool = True,
        prefix: str = "locked",
        check_interval: float = 0,
    ) -> Callable[[DecoratedFunc], DecoratedFunc]:
        return decorators.locked(
            backend=self,  # type: ignore[arg-type]
            ttl=ttl,
            key=key,
            wait=wait,
            prefix=prefix,
            check_interval=check_interval,
        )

    def bloom(
        self,
        *,
        capacity: int,
        name: KeyOrTemplate | None = None,
        false_positives: float | int = 1,
        check_false_positive: bool = True,
        prefix: str = "bloom",
    ) -> Callable[[DecoratedFunc], DecoratedFunc]:
        return decorators.bloom(
            backend=self,  # type: ignore[arg-type]
            name=name,
            false_positives=false_positives,
            capacity=capacity,
            check_false_positive=check_false_positive,
            prefix=prefix,
        )

    def dual_bloom(
        self,
        *,
        capacity: int,
        name: KeyOrTemplate | None = None,
        false: IntOrPair = 1,
        no_collisions: bool = False,
        prefix: str = "dual_bloom",
    ) -> Callable[[DecoratedFunc], DecoratedFunc]:
        return decorators.dual_bloom(
            backend=self,  # type: ignore[arg-type]
            name=name,
            false=false,
            no_collisions=no_collisions,
            capacity=capacity,
            prefix=prefix,
        )



from starlette.responses import StreamingResponse

from cashews.backends.interface import Backend
from cashews.serialize import DecodeError, register_type


async def encode_streaming_response(
    value: StreamingResponse, backend: Backend, key: str, expire: int, **kwargs
) -> bytes:
    value.body_iterator = set_iterator(backend, key, value.body_iterator, expire)
    serialized_value = b""
    serialized_value += bytes(str(value.status_code), "utf-8") + b":"
    for header, header_value in value.raw_headers:
        serialized_value += header + b"=" + header_value + b";"
    return serialized_value


async def decode_streaming_response(value: bytes, backend: Backend, key: str, **kwargs) -> StreamingResponse:
    if not await backend.get(f"{key}:done"):
        raise DecodeError()
    status_code, headers = value.split(b":", maxsplit=1)
    raw_headers = []
    for header in headers.split(b";", maxsplit=1):
        if not header:
            continue
        header_name, header_value = header.split(b"=", maxsplit=1)
        raw_headers.append((header_name, header_value))

    content = get_iterator(backend, key)
    resp = StreamingResponse(content=content, status_code=int(status_code))
    resp.raw_headers = raw_headers
    return resp


async def set_iterator(backend: Backend, key: str, iterator, expire: int):
    chunk_number = 0
    async for chunk in iterator:
        await backend.set(f"{key}:chunk:{chunk_number}", chunk, expire=expire)
        yield chunk
        chunk_number += 1
    await backend.set(f"{key}:done", True, expire=expire)  # mark as finished


async def get_iterator(backend: Backend, key: str):
    chunk_number = 0
    while True:
        chunk = await backend.get(f"{key}:chunk:{chunk_number}")
        if not chunk:
            return
        yield chunk
        chunk_number += 1


register_type(StreamingResponse, encode_streaming_response, decode_streaming_response)



class RaiseException:
    def __init__(self, exc: Exception):
        self.exc = exc


def return_or_raise(result):
    if isinstance(result, RaiseException):
        raise result.exc
    return result



import asyncio
import random

from cashews import Cache
from cashews.validation import invalidate_further


async def test_invalidate_decor_str(cache: Cache):
    @cache(ttl=1, key="key:{arg}")
    async def func(arg):
        return random.random()

    @cache.invalidate("key:{arg}")
    async def func2(arg, default=True):
        return random.random()

    first_call = await func("test")
    await asyncio.sleep(0.01)

    assert first_call == await func("test")
    await func2("test")
    await asyncio.sleep(0.01)
    assert first_call != await func("test")


async def test_invalidate_decor_tag(cache: Cache):
    @cache(ttl=1, key="key:{arg}", tags=("key:{arg}", "key"))
    async def func(arg):
        return random.random()

    @cache.invalidate("key:{arg}")
    async def func2(arg, default=True):
        return random.random()

    first_call = await func("test")
    await asyncio.sleep(0.01)

    assert first_call == await func("test")
    await func2("test")
    await asyncio.sleep(0.01)
    assert first_call != await func("test")


async def test_invalidate_decor_tag_func(cache: Cache):
    @cache(ttl=1, key="key:{arg:hash}", tags=("key:{arg}", "key"))
    async def func(arg):
        return random.random()

    @cache.invalidate("key:{arg:hash}")
    async def func2(arg, default=True):
        return random.random()

    first_call = await func("test")
    await asyncio.sleep(0.01)

    assert first_call == await func("test")
    await func2("test")
    await asyncio.sleep(0.01)
    assert first_call != await func("test")


async def test_invalidate_further_decorator(cache):
    @cache(ttl=100)
    async def func():
        return random.random()

    first = await func()
    assert await func() == first

    with invalidate_further():
        on_clear = await func()
    assert on_clear is not None
    assert on_clear != first  # the key was deleted
    second = await func()
    assert second != first
    assert await func() == second


async def test_invalidate_further_get(cache):
    await cache.set("key", "value")
    assert await cache.get("key") == "value"

    with invalidate_further():
        assert await cache.set("key", "value2")
        assert await cache.set("key2", "value2")
        assert await cache.get("key") is None

    await asyncio.sleep(0.01)
    assert await cache.get("key") is None
    assert await cache.get("key2") == "value2"
    assert await cache.set("key", "value3")
    assert await cache.get("key") == "value3"


async def test_invalidate_further_get_many(cache):
    await cache.set("key", "value")
    await cache.set("key2", "value2")

    assert await cache.get_many("key", "key2") == ("value", "value2")

    with invalidate_further():
        assert await cache.get_many("key") == (None,)

    await asyncio.sleep(0.01)
    assert await cache.get_many("key", "key2") == (None, "value2")


async def test_invalidate_further_get_match(cache):
    await cache.set("key1", "value1")
    await cache.set("key2", "value2")
    assert {k: v async for k, v in cache.get_match("key*")} == {
        "key1": "value1",
        "key2": "value2",
    }

    with invalidate_further():
        assert {k: v async for k, v in cache.get_match("key1*")} == {}

    await asyncio.sleep(0.01)
    assert {k: v async for k, v in cache.get_match("key*")} == {"key2": "value2"}



from unittest import mock

import pytest

from cashews import Cache
from cashews.helpers import memory_limit


@pytest.mark.parametrize(
    ("min_bytes", "max_bytes", "size", "called"),
    (
        (10, 80, 11, True),
        (10, 80, 10, True),
        (10, 80, 80, False),
        (1, 11, 10, False),
        (10, 1, 80, False),
    ),
)
async def test_memory_limit_set(cache: Cache, target, min_bytes, max_bytes, size, called):
    cache.add_middleware(memory_limit(min_bytes=min_bytes, max_bytes=max_bytes))

    await cache.set(key="key", value="v" * size)

    if called:
        target.set.assert_called_once_with(key="key", value=mock.ANY, expire=None, exist=None)
    else:
        target.set.assert_not_called()


async def test_memory_limit_set_many(cache: Cache, target):
    cache.add_middleware(memory_limit(min_bytes=52, max_bytes=75))

    await cache.set_many({"key": "v" * 35})
    target.set_many.assert_not_called()

    await cache.set_many({"key": "v" * 35, "key2": "v"})
    target.set_many.assert_not_called()

    await cache.set_many({"key": "v" * 35, "key2": "v", "key3": "v" * 15})
    target.set_many.assert_called_once_with(pairs={"key3": mock.ANY}, expire=None)



from __future__ import annotations

from contextvars import ContextVar, Token
from enum import Enum
from functools import wraps
from typing import TYPE_CHECKING

from cashews.backends.interface import Backend
from cashews.backends.transaction import LockTransactionBackend, TransactionBackend

from .wrapper import Wrapper

if TYPE_CHECKING:  # pragma: no cover
    from cashews._typing import DecoratedFunc

_transaction: ContextVar[Transaction | None] = ContextVar("transaction", default=None)


class TransactionMode(Enum):
    FAST = "fast"  # simple inmemory impl, allow to have lost updates,
    LOCKED = "locked"  # lock per key - not allow to work in parallel with the same key
    SERIALIZABLE = "serializable"  # global lock - not allow any parallel changes


class TransactionWrapper(Wrapper):
    transaction_mode = TransactionMode.LOCKED
    transaction_timeout = 10

    def set_transaction_timeout(self, timeout: int) -> None:
        self.transaction_timeout = timeout

    def set_transaction_mode(self, mode: TransactionMode) -> None:
        self.transaction_mode = mode

    def _get_backend(self, key: str) -> Backend:
        backend = super()._get_backend(key)
        tx: Transaction | None = _transaction.get()
        if tx:
            return tx.wrap(backend)
        return backend

    def transaction(
        self, mode: TransactionMode | None = None, timeout: float | None = None
    ) -> TransactionContextDecorator:
        mode = mode or self.transaction_mode
        timeout = timeout or self.transaction_timeout
        return TransactionContextDecorator(mode, timeout)


class TransactionContextDecorator:
    __slots__ = ["_mode", "_timeout", "_inner", "_return_token"]

    def __init__(self, mode: TransactionMode | None = None, timeout: float | None = None):
        self._mode = mode
        self._timeout = timeout
        self._inner = False
        self._return_token: Token | None = None

    @property
    def current_tx(self) -> Transaction | None:
        return _transaction.get()

    async def __aenter__(self) -> Transaction:
        if self.current_tx:
            self._inner = True
            return self.current_tx
        return self.start()

    def start(self) -> Transaction:
        tx = Transaction(self._mode, self._timeout)
        self._return_token = _transaction.set(tx)
        return tx

    def close(self):
        _transaction.reset(self._return_token)

    async def __aexit__(self, exc_type, exc_value, exc_tb) -> None:
        if not self.current_tx or self._inner:
            self._inner = False
            return
        if not exc_tb:
            await self.commit()
        else:
            await self.rollback()
        self.close()

    def __call__(self, func: DecoratedFunc) -> DecoratedFunc:
        @wraps(func)
        async def wrapper(*args, **kwargs):
            async with self:
                return await func(*args, **kwargs)

        return wrapper  # type: ignore[return-value]

    async def commit(self) -> None:
        if self.current_tx:
            await self.current_tx.commit()

    async def rollback(self) -> None:
        if self.current_tx:
            await self.current_tx.rollback()


class Transaction:
    __slots__ = ["_mode", "_timeout", "_backends"]

    def __init__(self, mode: TransactionMode | None = None, timeout: float | None = None):
        self._mode = mode
        self._timeout = timeout
        self._backends: dict[str, TransactionBackend] = {}

    def wrap(self, backend: Backend) -> Backend:
        if backend._id not in self._backends:
            self._backends[backend._id] = self._get_tx_backend(backend)
        return self._backends[backend._id]

    def _get_tx_backend(self, backend: Backend) -> TransactionBackend:
        if self._mode == TransactionMode.FAST:
            return TransactionBackend(backend)
        if self._mode == TransactionMode.SERIALIZABLE:
            return LockTransactionBackend(backend, serializable=True, timeout=self._timeout)
        return LockTransactionBackend(backend, serializable=False, timeout=self._timeout)

    async def commit(self) -> None:
        for tx_backend in list(self._backends.values()):
            await tx_backend.commit()

    async def rollback(self) -> None:
        for tx_backend in list(self._backends.values()):
            await tx_backend.rollback()



from __future__ import annotations

from collections.abc import Iterator
from contextlib import contextmanager, suppress
from typing import TYPE_CHECKING

from cashews.commands import Command
from cashews.exceptions import NotConfiguredError

from .wrapper import Wrapper

if TYPE_CHECKING:  # pragma: no cover
    from cashews._typing import AsyncCallable_T
    from cashews.backends.interface import Backend


async def _is_disable_middleware(call: AsyncCallable_T, cmd: Command, backend: Backend, *args, **kwargs):
    if backend.is_disable(cmd):
        if cmd in (Command.GET, Command.GET_MANY):
            return kwargs.get("default")
        return None
    return await call(*args, **kwargs)


class ControlWrapper(Wrapper):
    def __init__(self, name: str = ""):
        super().__init__(name)
        self.add_middleware(_is_disable_middleware)

    def disable(self, *cmds: Command, prefix: str = "") -> None:
        with suppress(NotConfiguredError):
            return self._get_backend(prefix).disable(*cmds)

    def enable(self, *cmds: Command, prefix: str = "") -> None:
        return self._get_backend(prefix).enable(*cmds)

    @contextmanager
    def disabling(self, *cmds: Command, prefix: str = "") -> Iterator[None]:
        self.disable(*cmds, prefix=prefix)
        try:
            yield
        finally:
            with suppress(NotConfiguredError):
                self.enable(*cmds, prefix=prefix)

    def is_disable(self, *cmds: Command, prefix: str = "") -> bool:
        return self._get_backend(prefix).is_disable(*cmds)

    def is_enable(self, *cmds: Command, prefix: str = "") -> bool:
        return not self.is_disable(*cmds, prefix=prefix)

    @property
    def is_full_disable(self) -> bool:
        return all(backend.is_full_disable for backend in self._backends.values())



from __future__ import annotations

import asyncio
import uuid
from abc import ABCMeta, abstractmethod
from collections.abc import AsyncGenerator, AsyncIterator, Iterable, Mapping
from contextlib import asynccontextmanager
from contextvars import ContextVar
from typing import TYPE_CHECKING, Any, overload

from cashews.commands import ALL, Command
from cashews.exceptions import CacheBackendInteractionError, LockedError
from cashews.serialize import Serializer

if TYPE_CHECKING:  # pragma: no cover
    from cashews._typing import Default, Key, OnRemoveCallback, Value

NOT_EXIST = -2
UNLIMITED = -1


class _BackendInterface(metaclass=ABCMeta):
    @property
    @abstractmethod
    def is_init(self) -> bool: ...

    @abstractmethod
    async def init(self): ...

    @abstractmethod
    async def close(self): ...

    @abstractmethod
    async def set(
        self,
        key: Key,
        value: Value,
        expire: float | None = None,
        exist: bool | None = None,
    ) -> bool: ...

    @abstractmethod
    async def set_many(self, pairs: Mapping[Key, Value], expire: float | None = None) -> None: ...

    @abstractmethod
    async def set_raw(self, key: Key, value: Value, **kwargs: Any) -> None: ...

    @overload
    async def get(self, key: Key, default: Default) -> Value | Default: ...

    @overload
    async def get(self, key: Key, default: None = None) -> Value | None: ...

    @abstractmethod
    async def get(self, key: Key, default: Default | None = None) -> Value | Default | None: ...

    @abstractmethod
    async def get_raw(self, key: Key) -> Value: ...

    @abstractmethod
    async def get_many(self, *keys: Key, default: Value | None = None) -> tuple[Value | None, ...]: ...

    @abstractmethod
    def get_match(self, pattern: str, batch_size: int = 100) -> AsyncIterator[tuple[Key, Value]]: ...

    @abstractmethod
    async def exists(self, key: Key) -> bool: ...

    @abstractmethod
    def scan(self, pattern: str, batch_size: int = 100) -> AsyncIterator[Key]: ...

    @abstractmethod
    async def incr(self, key: Key, value: int = 1, expire: float | None = None) -> int: ...

    @abstractmethod
    async def delete(self, key: Key) -> bool: ...

    @abstractmethod
    async def delete_many(self, *keys: Key) -> None: ...

    @abstractmethod
    async def delete_match(self, pattern: str) -> None: ...

    @abstractmethod
    async def expire(self, key: Key, timeout: float): ...

    @abstractmethod
    async def get_expire(self, key: Key) -> int: ...

    @abstractmethod
    async def get_bits(self, key: Key, *indexes: int, size: int = 1) -> tuple[int, ...]: ...

    @abstractmethod
    async def incr_bits(self, key: Key, *indexes: int, size: int = 1, by: int = 1) -> tuple[int, ...]: ...

    @abstractmethod
    async def slice_incr(
        self,
        key: Key,
        start: int | float,
        end: int | float,
        maxvalue: int,
        expire: float | None = None,
    ) -> int: ...

    @abstractmethod
    async def set_add(self, key: Key, *values: str, expire: float | None = None) -> None: ...

    @abstractmethod
    async def set_remove(self, key: Key, *values: str) -> None: ...

    @abstractmethod
    async def set_pop(self, key: Key, count: int = 100) -> Iterable[str]: ...

    @abstractmethod
    async def get_size(self, key: Key) -> int:
        """
        Return size in bites that allocated by a value for given key
        """
        ...

    @abstractmethod
    async def get_keys_count(self) -> int:
        """
        Return count keys in cache
        """
        ...

    @abstractmethod
    async def ping(self, message: bytes | None = None) -> bytes: ...

    @abstractmethod
    async def clear(self) -> None: ...

    async def set_lock(self, key: Key, value: Value, expire: float) -> bool:
        return await self.set(key, value, expire=expire, exist=False)

    @abstractmethod
    async def is_locked(
        self,
        key: Key,
        wait: float | None = None,
        step: float = 0.1,
    ) -> bool: ...

    @abstractmethod
    async def unlock(self, key: Key, value: Value) -> bool: ...

    @asynccontextmanager
    async def lock(
        self, key: Key, expire: float, wait: bool = True, check_interval: float = 0
    ) -> AsyncGenerator[None]:
        identifier = str(uuid.uuid4())
        while True:
            lock = await self.set_lock(key, identifier, expire=expire)
            if not lock:
                # we need to check the connection by ping because
                # if redis unavailable and a backend have flag `safe`
                # we will have a brake lock
                try:
                    if await self.ping(b"LOCK") is None:
                        yield
                        return
                except CacheBackendInteractionError:
                    yield
                    return

                if wait:
                    await asyncio.sleep(check_interval)
                    continue
                raise LockedError(f"Key {key} is already locked")
            try:
                yield
            finally:
                await self.unlock(key, identifier)
            return


class ControlMixin:
    enable_by_default = True

    def __init__(self, *args, **kwargs) -> None:
        self.__disable: ContextVar[set[Command]] = ContextVar(str(id(self)), default=set())
        self._control_set = False
        super().__init__(*args, **kwargs)

    @property
    def _disable(self) -> set[Command]:
        return self.__disable.get()

    def _set_disable(self, value: set[Command]) -> None:
        self.__disable.set(value)
        self._control_set = True

    def is_disable(self, *cmds: Command) -> bool:
        if not self._control_set:
            return not self.enable_by_default
        _disable = self._disable
        if not cmds and _disable:
            return True
        for cmd in cmds:
            if cmd in _disable:
                return True
        return False

    def is_enable(self, *cmds: Command) -> bool:
        return not self.is_disable(*cmds)

    @property
    def is_full_disable(self) -> bool:
        if not self._control_set:
            return not self.enable_by_default
        return self._disable == ALL

    def disable(self, *cmds: Command) -> None:
        if not cmds:
            _disable = ALL.copy()
        else:
            _disable = self._disable.copy()
            _disable.update(cmds)
        self._set_disable(_disable)

    def enable(self, *cmds: Command) -> None:
        if not cmds:
            _disable = set()
        else:
            _disable = self._disable.copy()
            _disable -= set(cmds)
        self._set_disable(_disable)


class Backend(ControlMixin, _BackendInterface, metaclass=ABCMeta):
    def __init__(self, *args, serializer: Serializer | None = None, **kwargs) -> None:
        super().__init__()
        self._id = uuid.uuid4().hex
        self._serializer = serializer
        self._on_remove_callbacks: list[OnRemoveCallback] = []

    def on_remove_callback(self, callback: OnRemoveCallback) -> None:
        self._on_remove_callbacks.append(callback)

    async def _call_on_remove_callbacks(self, *keys: Key) -> None:
        for callback in self._on_remove_callbacks:
            await callback(keys, backend=self)



from __future__ import annotations

from functools import partial
from typing import Any

from ._typing import CacheCondition, CallableCacheCondition


def _not_none_store_condition(result, args, kwargs, key=None) -> bool:
    return result is not None


def _store_all(result, args, kwargs, key=None) -> bool:
    return True


def _exceptions(*exceptions: type[Exception], default: bool = True) -> CacheCondition:
    exceptions = exceptions or Exception  # type: ignore

    def _cond(result: Any, args: tuple, kwargs: dict[str, Any], key: str = "") -> bool | Exception:
        if isinstance(result, exceptions):
            return result
        return default

    return _cond


with_exceptions = partial(_exceptions, default=True)
only_exceptions = partial(_exceptions, default=False)


NOT_NONE = "not_none"
_ALL_CONDITIONS = {Any, "all", any, "any", None}
_NOT_NONE_CONDITIONS = {NOT_NONE, "skip_none"}


def get_cache_condition(condition: CacheCondition) -> CallableCacheCondition:
    if condition in _ALL_CONDITIONS:
        return _store_all
    if condition in _NOT_NONE_CONDITIONS:
        return _not_none_store_condition
    return condition  # type: ignore[return-value]



from __future__ import annotations

import inspect
from collections.abc import AsyncIterator, Iterable, Mapping
from functools import partial
from typing import TYPE_CHECKING, overload

from cashews.backends.interface import Backend
from cashews.commands import Command
from cashews.ttl import ttl_to_seconds

from .wrapper import Wrapper

if TYPE_CHECKING:  # pragma: no cover
    from cashews._typing import TTL, AsyncCallable_T, Callable_T, Default, Key, Result_T, Value

_empty = object()


class CommandWrapper(Wrapper):
    async def set(
        self,
        key: Key,
        value: Value,
        expire: TTL = None,
        exist: bool | None = None,
    ) -> bool:
        return await self._with_middlewares(Command.SET, key)(
            key=key,
            value=value,
            expire=ttl_to_seconds(expire),
            exist=exist,
        )

    async def set_raw(self, key: Key, value: Value, **kwargs) -> None:
        return await self._with_middlewares(Command.SET_RAW, key)(key=key, value=value, **kwargs)

    @overload
    async def get(self, key: Key, default: Default) -> Value | Default: ...

    @overload
    async def get(self, key: Key, default: None = None) -> Value | None: ...

    async def get(self, key: Key, default: Default | None = None) -> Value | Default | None:
        return await self._with_middlewares(Command.GET, key)(key=key, default=default)

    async def get_or_set(
        self, key: Key, default: Default | AsyncCallable_T | Callable_T, expire: TTL = None
    ) -> Value | Default | Result_T:
        value = await self.get(key, default=_empty)
        if value is not _empty:
            return value
        if inspect.isawaitable(default):
            _default = await default
        elif callable(default):
            if inspect.iscoroutinefunction(default):
                _default = await default()
            else:
                _default = default()
        else:
            _default = default
        await self.set(key, _default, expire=expire)
        return _default

    async def get_raw(self, key: Key) -> Value:
        return await self._with_middlewares(Command.GET_RAW, key)(key=key)

    async def scan(self, pattern: str, batch_size: int = 100) -> AsyncIterator[Key]:
        backend = self._get_backend(pattern)

        async def call(pattern, batch_size):
            return backend.scan(pattern=pattern, batch_size=batch_size)

        for middleware in reversed(self._default_middlewares):
            call = partial(middleware, call, Command.SCAN, backend)

        for middleware in self._middlewares[backend._id]:
            call = partial(middleware, call, Command.SCAN, backend)

        async for key in await call(pattern=pattern, batch_size=batch_size):
            yield key

    async def get_match(
        self,
        pattern: str,
        batch_size: int = 100,
    ) -> AsyncIterator[tuple[Key, Value]]:
        backend = self._get_backend(pattern)
        middlewares = self._middlewares[backend._id]

        async def call(pattern, batch_size):
            return backend.get_match(pattern=pattern, batch_size=batch_size)

        for middleware in reversed(self._default_middlewares):
            call = partial(middleware, call, Command.GET_MATCH, backend)

        for middleware in middlewares:
            call = partial(middleware, call, Command.GET_MATCH, backend)
        async for key, value in await call(pattern=pattern, batch_size=batch_size):
            yield key, value

    async def get_many(self, *keys: Key, default: Value | None = None) -> tuple[Value | None, ...]:
        backends: dict[Backend, list[str]] = {}
        for key in keys:
            backend = self._get_backend(key)
            backends.setdefault(backend, []).append(key)
        result: dict[Key, Value] = {}
        for _keys in backends.values():
            _values = await self._with_middlewares(Command.GET_MANY, _keys[0])(*_keys, default=default)
            result.update(dict(zip(_keys, _values)))
        return tuple(result.get(key) for key in keys)

    async def set_many(self, pairs: Mapping[Key, Value], expire: TTL = None):
        backends: dict[Backend, list[Key]] = {}
        for key in pairs:
            backend = self._get_backend(key)
            backends.setdefault(backend, []).append(key)
        for keys in backends.values():
            data = {key: pairs[key] for key in keys}
            await self._with_middlewares(Command.SET_MANY, keys[0])(
                pairs=data,
                expire=ttl_to_seconds(expire),
            )

    async def get_bits(self, key: Key, *indexes: int, size: int = 1) -> tuple[int, ...]:
        return await self._with_middlewares(Command.GET_BITS, key)(key, *indexes, size=size)

    async def incr_bits(self, key: Key, *indexes: int, size: int = 1, by: int = 1) -> tuple[int, ...]:
        return await self._with_middlewares(Command.INCR_BITS, key)(key, *indexes, size=size, by=by)

    async def slice_incr(
        self,
        key: Key,
        start: int | float,
        end: int | float,
        maxvalue: int,
        expire: TTL = None,
    ) -> int:
        return await self._with_middlewares(Command.SLICE_INCR, key)(
            key=key,
            start=start,
            end=end,
            maxvalue=maxvalue,
            expire=ttl_to_seconds(expire),
        )

    async def incr(self, key: Key, value: int = 1, expire: float | None = None) -> int:
        return await self._with_middlewares(Command.INCR, key)(key=key, value=value, expire=expire)

    async def delete(self, key: Key) -> bool:
        return await self._with_middlewares(Command.DELETE, key)(key=key)

    async def delete_many(self, *keys: Key) -> None:
        backends: dict[Backend, list[Key]] = {}
        for key in keys:
            backend = self._get_backend(key)
            backends.setdefault(backend, []).append(key)
        for _keys in backends.values():
            await self._with_middlewares(Command.DELETE_MANY, _keys[0])(*_keys)

    async def delete_match(self, pattern: str) -> None:
        return await self._with_middlewares(Command.DELETE_MATCH, pattern)(pattern=pattern)

    async def expire(self, key: Key, timeout: TTL):
        return await self._with_middlewares(Command.EXPIRE, key)(key=key, timeout=ttl_to_seconds(timeout))

    async def get_expire(self, key: Key) -> int:
        return await self._with_middlewares(Command.GET_EXPIRE, key)(key=key)

    async def exists(self, key: Key) -> bool:
        return await self._with_middlewares(Command.EXISTS, key)(key=key)

    async def set_lock(self, key: Key, value: Value, expire: TTL) -> bool:
        return await self._with_middlewares(Command.SET_LOCK, key)(key=key, value=value, expire=ttl_to_seconds(expire))

    async def unlock(self, key: Key, value: Value) -> bool:
        return await self._with_middlewares(Command.UNLOCK, key)(key=key, value=value)

    async def get_size(self, key: Key) -> int:
        return await self._with_middlewares(Command.GET_SIZE, key)(key=key)

    async def ping(self, message: bytes | None = None) -> bytes:
        message = b"PING" if message is None else message
        return await self._with_middlewares(Command.PING, message.decode())(message=message)

    async def get_keys_count(self) -> int:
        result = 0
        for backend in self._backends.values():
            count = await self._with_middlewares_for_backend(
                Command.GET_KEYS_COUNT, backend, self._default_middlewares
            )()
            result += count
        return result

    async def clear(self) -> None:
        for backend in self._backends.values():
            await self._with_middlewares_for_backend(Command.CLEAR, backend, self._default_middlewares)()

    async def is_locked(
        self,
        key: Key,
        wait: TTL = None,
        step: int | float = 0.1,
    ) -> bool:
        return await self._with_middlewares(Command.IS_LOCKED, key)(key=key, wait=ttl_to_seconds(wait), step=step)

    async def set_add(self, key: Key, *values: str, expire: TTL = None) -> None:
        return await self._with_middlewares(Command.SET_ADD, key)(key, *values, expire=ttl_to_seconds(expire))

    async def set_remove(self, key: Key, *values: str) -> None:
        return await self._with_middlewares(Command.SET_REMOVE, key)(key, *values)

    async def set_pop(self, key: Key, count: int = 100) -> Iterable[str]:
        return await self._with_middlewares(Command.SET_POP, key)(key=key, count=count)



from __future__ import annotations

from collections.abc import Iterable
from functools import lru_cache
from re import Match, Pattern

from cashews._typing import TTL, Key, KeyOrTemplate, OnRemoveCallback, Tag, Tags, Value
from cashews.backends.interface import Backend
from cashews.formatter import default_format, template_to_re_pattern

from .commands import CommandWrapper


class TagsRegistry:
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._registry_template: dict[Tag, list[Pattern]] = {}

    def register_tag(self, tag: Tag, key_template: KeyOrTemplate):
        self._registry_template.setdefault(tag, [])
        self._registry_template[tag].append(template_to_re_pattern(key_template))

    def get_key_tags(self, key: Key) -> Tags:
        tags = []
        for tag, patterns in self._registry_template.items():
            match = self._match_patterns(key, patterns)
            if match:
                group_dict = {k: v if v is not None else "" for k, v in match.groupdict().items()}
                tag = default_format(tag, **group_dict)
                tags.append(tag)
        return tags

    @staticmethod
    def _match_patterns(key: Key, patterns: list[Pattern]) -> Match | None:
        for pattern in patterns:
            match = pattern.fullmatch(key)
            if match:
                return match
        return None


class CommandsTagsWrapper(CommandWrapper):
    def __init__(self, name: str = ""):
        super().__init__(name)
        self._tags_registry = TagsRegistry()
        self._tags_key_prefix = "_tag:"
        self._on_remove_cb = self._on_remove_callback()

    def setup_tags_backend(self, settings_url: str, middlewares: tuple = (), **kwargs) -> Backend:
        return self.setup(
            settings_url,
            middlewares=middlewares,
            prefix=self._tags_key_prefix,
            **kwargs,
        )

    @lru_cache(maxsize=1)  # noqa: B019
    def _get_tags_backend(self):
        return self._get_backend(self._tags_key_prefix)

    @property
    def tags_backend(self):
        return self._get_tags_backend()

    def register_tag(self, tag: Tag, key_template: KeyOrTemplate):
        self._tags_registry.register_tag(tag, key_template)

    def _add_backend(self, backend: Backend, *args, **kwargs):
        super()._add_backend(backend, *args, **kwargs)
        backend.on_remove_callback(self._on_remove_cb)

    def _on_remove_callback(self) -> OnRemoveCallback:
        async def _callback(keys: Iterable[Key], backend: Backend) -> None:
            for tag, _keys in self._group_by_tags(keys).items():
                await self.tags_backend.set_remove(self._tags_key_prefix + tag, *_keys)

        return _callback

    def _group_by_tags(self, keys: Iterable[Key]) -> dict[Tag, list[Key]]:
        tags: dict[Tag, list[Key]] = {}
        for key in keys:
            for tag in self.get_key_tags(key):
                tags.setdefault(tag, []).append(key)
        return tags

    def get_key_tags(self, key: Key) -> Tags:
        return self._tags_registry.get_key_tags(key)

    async def delete_tags(self, *tags: Tag):
        for tag in tags:
            await self._delete_tag(tag)

    async def _delete_tag(self, tag: Tag):
        while True:
            keys = await self.set_pop(key=self._tags_key_prefix + tag, count=100)
            if not keys:
                break
            keys = list(keys)
            await self.delete_many(*keys)
            if len(keys) != 100:
                break

    async def set(
        self,
        key: Key,
        value: Value,
        expire: TTL = None,
        exist: bool | None = None,
        tags: Tags = (),
    ) -> bool:
        _set = await super().set(key=key, value=value, expire=expire, exist=exist)
        if _set and tags:
            for tag in tags:
                await self.set_add(self._tags_key_prefix + tag, key, expire=expire)
        return _set

    async def incr(self, key: Key, value: int = 1, expire: float | None = None, tags: Tags = ()) -> int:
        _set = await super().incr(key=key, value=value, expire=expire)
        if _set and tags:
            for tag in tags:
                await self.set_add(self._tags_key_prefix + tag, key, expire=expire)
        return _set



import asyncio
from unittest.mock import Mock

import pytest

from cashews import Cache, decorators

EXPIRE = 0.02


class CustomError(Exception):
    pass


async def test_fail_cache_simple(cache: Cache):
    @cache.failover(ttl=EXPIRE, exceptions=CustomError, key="fail")
    async def func(fail=False):
        if fail:
            raise CustomError()
        return b"ok"

    assert await func() == b"ok"
    assert await func(fail=True) == b"ok"

    await asyncio.sleep(EXPIRE * 2)
    with pytest.raises(CustomError):
        await func(fail=True)


async def test_fail_cache_fast_condition(cache: Cache):
    mem = set()

    def getter(key):
        return key in mem

    def setter(key, _):
        mem.add(key)

    fast_condition = decorators.fast_condition(getter=getter, setter=setter)

    @cache.failover(ttl=EXPIRE, condition=fast_condition, key="fail", prefix="")
    async def func(fail=False):
        if fail:
            raise CustomError()
        return b"ok"

    assert await func() == b"ok"
    assert await func(fail=True) == b"ok"
    assert await func(fail=True) == b"ok"
    assert "fail" in mem


async def test_cache_simple(cache: Cache):
    @cache.cache(ttl=EXPIRE, key="key")
    async def func(resp=b"ok"):
        return resp

    assert await func() == b"ok"

    await asyncio.sleep(0)
    assert await func(b"notok") == b"ok"

    await asyncio.sleep(EXPIRE * 1.1)
    assert await func(b"notok") == b"notok"


async def test_cache_simple_none(cache: Cache):
    mock = Mock()

    @cache(ttl=EXPIRE, key="key")
    async def func():
        mock()
        return None

    assert await func() is None
    assert mock.call_count == 1

    assert await func() is None
    assert mock.call_count == 1


async def test_cache_simple_cond(cache: Cache):
    mock = Mock()

    @cache(ttl=EXPIRE, key="key", condition=lambda x, *args, **kwargs: x == b"hit")
    async def func(resp=b"ok"):
        mock()
        return resp

    await func()
    assert mock.call_count == 1

    await func(b"notok")
    assert mock.call_count == 2

    await func(b"hit")
    assert mock.call_count == 3
    await func(b"hit")
    assert mock.call_count == 3


async def test_cache_simple_ttl(cache: Cache):
    mock = Mock()

    def _ttl(resp=b"ok"):
        if resp == b"ok":
            return 0.01
        return "2h"

    @cache(ttl=_ttl)
    async def func(resp=b"ok"):
        mock()
        return resp

    await func()
    await func()

    assert mock.call_count == 1
    await asyncio.sleep(0.02)

    await func()
    assert mock.call_count == 2

    await func(b"notok")
    assert mock.call_count == 3

    await asyncio.sleep(0.02)
    await func(b"notok")
    assert mock.call_count == 3


async def test_cache_simple_ttl_with_result(cache: Cache):
    mock = Mock()

    def _ttl(result, resp=b"ok"):
        if result == b"ok":
            return 0.01
        return "2h"

    @cache(ttl=_ttl)
    async def func(resp=b"ok"):
        mock()
        return resp

    await func()
    await func()

    assert mock.call_count == 1
    await asyncio.sleep(0.02)

    await func()
    assert mock.call_count == 2

    await func(b"notok")
    await func(resp=b"notok")
    assert mock.call_count == 3

    await asyncio.sleep(0.02)
    await func(b"notok")
    assert mock.call_count == 3


async def test_early_cache_simple(cache: Cache):
    @cache.early(ttl=EXPIRE, key="key")
    async def func(resp=b"ok"):
        return resp

    assert await func() == b"ok"

    await asyncio.sleep(0)
    assert await func(b"notok") == b"ok"

    await asyncio.sleep(EXPIRE * 1.1)
    assert await func(b"notok") == b"notok"


async def test_early_cache_no_background(cache: Cache):
    mock = Mock()

    @cache.early(ttl=EXPIRE, early_ttl=0.01, key="key", background=False)
    async def func(resp=b"ok"):
        mock()
        return resp

    assert await func() == b"ok"
    assert mock.call_count == 1

    await asyncio.sleep(0)
    assert await func(b"notok") == b"ok"
    assert mock.call_count == 1

    await asyncio.sleep(EXPIRE * 1.1)
    assert await func(b"notok") == b"notok"
    assert mock.call_count == 2


async def test_soft_cache_simple(cache: Cache):
    mock = Mock()

    @cache.soft(ttl=4 * EXPIRE, soft_ttl=EXPIRE, key="key")
    async def func(resp=b"ok"):
        mock()
        return resp

    assert await func() == b"ok"
    assert mock.call_count == 1

    await asyncio.sleep(0)
    assert await func(b"notok") == b"ok"
    assert mock.call_count == 1

    await asyncio.sleep(EXPIRE * 1.1)
    assert await func(b"notok") == b"notok"
    assert mock.call_count == 2


async def test_soft_cache_on_exc(cache: Cache):
    mock = Mock()

    @cache.soft(ttl=4 * EXPIRE, soft_ttl=EXPIRE, key="key")
    async def func(resp=b"ok"):
        if resp == b"notok":
            raise ValueError()
        mock()
        return resp

    assert await func() == b"ok"
    assert mock.call_count == 1

    await asyncio.sleep(0)
    assert await func(b"notok") == b"ok"
    assert mock.call_count == 1

    await asyncio.sleep(EXPIRE * 1.1)

    assert await func(b"notok") == b"ok"
    assert mock.call_count == 1

    await asyncio.sleep(EXPIRE * 3)
    with pytest.raises(ValueError):
        await func(b"notok")


@pytest.mark.xfail
async def test_early_cache_parallel(cache: Cache):
    mock = Mock()

    @cache.early(ttl=0.1, early_ttl=0.05, key="key")
    async def func(resp=b"ok"):
        await asyncio.sleep(0.01)
        mock(resp)
        return resp

    assert await func() == b"ok"  # warm

    assert mock.call_count == 1

    for _ in range(8):
        await asyncio.sleep(0.01)
        await asyncio.gather(*[func() for _ in range(10)])

    assert mock.call_count in (2, 3, 4)  # depends on backend


async def test_hit_cache(cache: Cache):
    mock = Mock()

    @cache.hit(ttl=1000, cache_hits=10, key="test")
    async def func(resp=b"ok"):
        mock(resp)
        return resp

    await func()  # cache
    await asyncio.gather(*[func() for _ in range(10)])  # get 10 hits
    assert mock.call_count in [1, 2]
    await func()  # cache
    assert mock.call_count in [2, 3]

    await asyncio.gather(*[func() for _ in range(10)])
    assert mock.call_count in [2, 3, 4]


async def test_hit_cache_early(cache: Cache):
    mock = Mock()

    @cache.hit(ttl=10, cache_hits=5, key="test", update_after=1)
    async def func(resp=b"ok"):
        mock(resp)
        return resp

    assert await func(b"1") == b"1"  # nocache
    assert mock.call_count == 1

    assert await func(b"2") == b"1"  # cache and update
    assert mock.call_count == 1

    await asyncio.sleep(0.01)
    assert mock.call_count == 2
    assert await func(b"3") == b"2"  # cache from prev and also update
    await asyncio.sleep(0.01)
    assert mock.call_count == 3


async def test_context_cache_detect_simple(cache: Cache):
    @cache(ttl=EXPIRE, key="key")
    async def func(resp=b"ok"):
        return resp

    with decorators.context_cache_detect as detector:
        assert await func() == b"ok"
        assert detector.calls == {}

        await asyncio.sleep(0)
        assert await func(b"notok") == b"ok"
        assert list(detector.calls.keys()) == [
            "key",
        ]

        await asyncio.sleep(EXPIRE * 1.1)
        assert await func(b"notok") == b"notok"
        assert len(detector.calls) == 1

    assert decorators.context_cache_detect._levels == {}


async def test_context_cache_detect_deep(cache: Cache):
    @cache(ttl=EXPIRE, key="key1")
    async def func1():
        return 1

    @cache(ttl=EXPIRE, key="key2")
    async def func2():
        return 2

    async def func():
        return await asyncio.gather(func1(), func2())

    with decorators.context_cache_detect as detector:
        await func()
        assert detector.calls == {}

        await asyncio.sleep(0)
        await func()

        assert len(detector.calls) == 2
        assert "key1" in detector.calls
        assert "key2" in detector.calls
    assert decorators.context_cache_detect._levels == {}


async def test_context_cache_detect_context(cache: Cache):
    @cache(ttl=EXPIRE, key="key1")
    async def func1():
        return 1

    @cache(ttl=EXPIRE, key="key2")
    async def func2():
        return 2

    async def func(*funcs):
        with decorators.context_cache_detect as detector:
            await asyncio.gather(*funcs)
            return len(detector.calls)

    await cache.set("key1", "test")
    await cache.set("key2", "test")
    assert await func1() == "test"
    assert await func2() == "test"

    with decorators.context_cache_detect as detector:
        assert await asyncio.create_task(func(func1())) == 1
        assert await asyncio.create_task(func(func1(), func2())) == 2
        assert await asyncio.create_task(func(func2())) == 1
        assert await asyncio.create_task(func(func2(), func2())) == 1

        assert len(detector.calls) == 2

    assert decorators.context_cache_detect._levels == {}



from __future__ import annotations

import asyncio
import inspect
from functools import partial, wraps
from typing import TYPE_CHECKING, Callable

from cashews.backends.interface import _BackendInterface
from cashews.key import get_cache_key, get_cache_key_template
from cashews.ttl import ttl_to_seconds

if TYPE_CHECKING:  # pragma: no cover
    from cashews._typing import TTL, DecoratedFunc, Key, KeyOrTemplate

__all__ = ("locked",)


def locked(
    backend: _BackendInterface,
    key: KeyOrTemplate | None = None,
    ttl: TTL | None = None,
    wait: bool = True,
    prefix: str = "lock",
    check_interval: float = 0,
) -> Callable[[DecoratedFunc], DecoratedFunc]:
    """
    Decorator that can help you to solve Cache stampede problem (https://en.wikipedia.org/wiki/Cache_stampede),
    Lock following function calls till first one will be finished
    Can guarantee that one function call for given ttl, if ttl is None

    :param backend: cache backend
    :param key: custom cache key, may contain alias to args or kwargs passed to a call
    :param ttl: duration to lock wrapped function call
    :param wait: if true - wait till lock is released
    :param prefix: custom prefix for key, default 'lock'
    :param check_interval: interval in seconds between lock checks while it is waiting for the lock to be released
    """
    ttl = ttl_to_seconds(ttl)

    def _decor(func):
        _key_template = get_cache_key_template(func, key=key, prefix=prefix)
        if inspect.isasyncgenfunction(func):
            return _asyncgen_lock(func, backend, ttl, _key_template, wait, check_interval)
        return _coroutine_lock(func, backend, ttl, _key_template, wait, check_interval)

    return _decor


def _coroutine_lock(
    func: DecoratedFunc,
    backend: _BackendInterface,
    ttl: TTL | None,
    key_template: KeyOrTemplate,
    wait: bool,
    check_interval: float,
) -> DecoratedFunc:
    @wraps(func)
    async def _wrap(*args, **kwargs):
        _ttl = ttl_to_seconds(ttl, *args, **kwargs, with_callable=True)
        _cache_key = get_cache_key(func, key_template, args, kwargs)
        async with backend.lock(_cache_key, _ttl, wait=wait, check_interval=check_interval):
            return await func(*args, **kwargs)

    return _wrap  # type: ignore[return-value]


def _asyncgen_lock(
    func,
    backend: _BackendInterface,
    ttl: TTL | None,
    key_template: KeyOrTemplate,
    wait: bool,
    check_interval: float,
):
    @wraps(func)
    async def _wrap(*args, **kwargs):
        _ttl = ttl_to_seconds(ttl, *args, **kwargs, with_callable=True)
        _cache_key = get_cache_key(func, key_template, args, kwargs)
        async with backend.lock(_cache_key, _ttl, wait=wait, check_interval=check_interval):
            async for chunk in func(*args, **kwargs):
                yield chunk
            return

    return _wrap


def thunder_protection(
    key: KeyOrTemplate | None = None,
) -> Callable[[DecoratedFunc], DecoratedFunc]:
    tasks: dict[str, asyncio.Task] = {}

    def _decor(func: DecoratedFunc) -> DecoratedFunc:
        _key_template = get_cache_key_template(func, key=key)

        def done_callback(_key: Key, _: asyncio.Task):
            del tasks[_key]

        @wraps(func)
        async def _wrapper(*args, **kwargs):
            _key = get_cache_key(func, _key_template, args, kwargs)
            if _key in tasks:
                return await tasks[_key]
            task = asyncio.create_task(func(*args, **kwargs))
            tasks[_key] = task
            task.add_done_callback(partial(done_callback, _key))
            return await task

        return _wrapper  # type: ignore[return-value]

    return _decor



import asyncio

from cashews import Cache
from cashews.commands import Command


async def test_issue_382():
    cache = Cache()
    cache.setup("mem://")

    @cache(ttl=60, key="item:{item_id}")
    async def get_item(item_id: int):
        return f"item_{item_id}"

    def sync_get_item(item_id: int):
        return asyncio.run(get_item(item_id))

    await asyncio.get_running_loop().run_in_executor(None, sync_get_item, 123)


async def test_control_mixin_cross_context():
    """Test that ControlMixin's ContextVar works across different async contexts.

    This simulates the FastAPI lifespan pattern where cache.setup() is called in one
    context (lifespan) and cache operations are called in different contexts (HTTP requests).
    """
    cache = Cache()

    # Setup cache in one context (simulating FastAPI lifespan)
    cache.setup("mem://")
    cache.enable()  # This calls self.__disable.get() internally

    @cache(ttl=60, key="test:{value}")
    async def cached_func(value: int):
        return f"result_{value}"

    # Test in a different async context (simulating HTTP request)
    async def test_in_new_context():
        # This should not raise LookupError when checking if cache is enabled/disabled
        result = await cached_func(123)
        assert result == "result_123"

        # Test disable/enable in new context
        cache.disable(Command.SET)
        assert cache.is_disable(Command.SET)

        cache.enable(Command.SET)
        assert cache.is_enable(Command.SET)

        return result

    def sync_runner():
        return asyncio.run(test_in_new_context())

    # Run in executor with new event loop (different context)
    result = await asyncio.get_running_loop().run_in_executor(None, sync_runner)
    assert result == "result_123"



from cashews.cache_condition import NOT_NONE

from .bloom import bloom, dual_bloom
from .cache.defaults import CacheDetect, context_cache_detect
from .cache.early import early
from .cache.fail import failover, fast_condition
from .cache.hit import hit
from .cache.iterator import iterator
from .cache.simple import cache
from .cache.soft import soft
from .circuit_breaker import circuit_breaker
from .locked import locked, thunder_protection
from .rate import rate_limit
from .rate_slide import slice_rate_limit

__all__ = [
    "NOT_NONE",
    "bloom",
    "dual_bloom",
    "CacheDetect",
    "context_cache_detect",
    "early",
    "failover",
    "fast_condition",
    "hit",
    "iterator",
    "cache",
    "soft",
    "circuit_breaker",
    "locked",
    "thunder_protection",
    "rate_limit",
    "slice_rate_limit",
]



from __future__ import annotations

import asyncio
import re
from collections.abc import AsyncIterator, Iterable, Mapping
from datetime import datetime, timezone
from typing import Any

from diskcache import Cache, FanoutCache

from cashews._typing import Key, Value
from cashews.serialize import DEFAULT_SERIALIZER, Serializer
from cashews.utils import Bitarray

from .interface import NOT_EXIST, UNLIMITED, Backend


class DiskCache(Backend):
    def __init__(self, *args, directory=None, shards=8, **kwargs: Any) -> None:
        serializer = kwargs.pop("serializer", DEFAULT_SERIALIZER)
        self.__is_init = False
        self._set_locks: dict[str, asyncio.Lock] = {}
        self._sharded = shards > 1
        if not self._sharded:
            self._cache = Cache(directory=directory, **kwargs)
        else:
            self._cache = FanoutCache(directory=directory, shards=shards, **kwargs)
        super().__init__(serializer=serializer, **kwargs)
        self._serializer: Serializer

    async def init(self):
        self.__is_init = True

    async def _run_in_executor(self, call, *args):
        return await asyncio.get_running_loop().run_in_executor(None, call, *args)

    @property
    def is_init(self) -> bool:
        return self.__is_init

    async def close(self):
        self._cache.close()
        self.__is_init = False

    async def set(
        self,
        key: Key,
        value: Value,
        expire: float | None = None,
        exist: bool | None = None,
    ) -> bool:
        value = await self._serializer.encode(self, key=key, value=value, expire=expire)
        future = self._run_in_executor(self._set, key, value, expire, exist)
        if exist is not None:
            # we should have async lock until value real set
            lock = self._set_locks.setdefault(key, asyncio.Lock())
            async with lock:
                try:
                    return await future
                finally:
                    self._set_locks.pop(key, None)
        return await future

    def _set(self, key: Key, value: Value, expire=None, exist=None):
        if exist is not None and self._exists(key) is not exist:
            return False
        if expire is None:
            expire = self._get_expire(key)
            expire = expire if expire not in [UNLIMITED, NOT_EXIST] else None
        return self._cache.set(key, value, expire)

    async def set_raw(self, key: Key, value: Any, **kwargs: Any):
        return self._cache.set(key, value, **kwargs)

    async def get(self, key: Key, default: Value | None = None) -> Value:
        value = await self._run_in_executor(self._cache.get, key, default)
        return await self._serializer.decode(self, key=key, value=value, default=default)

    async def get_raw(self, key: Key) -> Value:
        return self._cache.get(key)

    async def get_many(self, *keys: Key, default: Value | None = None) -> tuple[Value | None, ...]:
        if not keys:
            return ()
        values = await self._run_in_executor(self._get_many, keys, default)
        values = await asyncio.gather(
            *[self._serializer.decode(self, key=key, value=value, default=default) for key, value in zip(keys, values)]
        )
        return tuple(None if isinstance(value, Bitarray) else value for value in values)

    def _get_many(self, keys: list[Key], default: Value | None = None):
        values = []
        for key in keys:
            val = self._cache.get(key, default=default)
            values.append(val)
        return values

    async def set_many(self, pairs: Mapping[Key, Value], expire: float | None = None):
        _pairs = {}
        for key, value in pairs.items():
            value = await self._serializer.encode(self, key=key, value=value, expire=expire)
            _pairs[key] = value
        return await self._run_in_executor(self._set_many, _pairs, expire)

    def _set_many(self, pairs: Mapping[Key, Value], expire: float | None = None):
        for key, value in pairs.items():
            self._set(key, value, expire=expire)

    async def exists(self, key: Key) -> bool:
        return await self._run_in_executor(self._exists, key)

    def _exists(self, key: Key) -> bool:
        return key in self._cache

    async def scan(self, pattern: str, batch_size: int = 100) -> AsyncIterator[Value]:
        if not self._sharded:
            for key in await self._run_in_executor(self._scan, pattern):
                yield key

    def _scan(self, pattern: str) -> Value:
        pattern = pattern.replace("*", ".*")
        regexp = re.compile(pattern)
        for key in self._cache.iterkeys():
            if regexp.fullmatch(key):
                yield key

    async def get_bits(self, key: Key, *indexes: int, size: int = 1) -> tuple[int, ...]:
        array = await self.get(key, default=Bitarray("0"))
        result = []
        for index in indexes:
            result.append(array.get(index, size))
        return tuple(result)

    async def incr_bits(self, key: str, *indexes: int, size: int = 1, by: int = 1) -> tuple[int, ...]:
        array = await self.get(key, default=Bitarray("0"))
        result = []
        for index in indexes:
            array.incr(index, size, by)
            result.append(array.get(index, size))
        await self.set(key, array)
        return tuple(result)

    async def incr(self, key: Key, value: int = 1, expire: float | None = None) -> int:
        return await self._run_in_executor(self._incr, key, value, expire)

    def _incr(self, key: Key, value: int = 1, expire: float | None = None) -> int:
        res = self._cache.incr(key, delta=value, retry=True)
        if res == 1 and expire:
            self._cache.touch(key, expire)
        return res

    async def delete(self, key: Key) -> bool:
        try:
            return await self._run_in_executor(self._cache.delete, key)
        finally:
            await self._call_on_remove_callbacks(key)

    async def delete_many(self, *keys: Key):
        try:
            await self._run_in_executor(self._delete_many, keys)
        finally:
            await self._call_on_remove_callbacks(*keys)

    def _delete_many(self, keys: list[Key]):
        for key in keys:
            self._cache.delete(key)

    async def delete_match(self, pattern: str):
        keys = []
        async for key in self.scan(pattern):
            keys.append(key)
        try:
            return await self._run_in_executor(self._delete_match, pattern)
        finally:
            await self._call_on_remove_callbacks(*keys)

    def _delete_match(self, pattern: str):
        for key in self._scan(pattern):
            self._cache.delete(key)

    async def get_match(self, pattern: str, batch_size: int = 100) -> AsyncIterator[tuple[Key, Value]]:
        if self._sharded:
            return
        for key in await self._run_in_executor(self._scan, pattern):
            value = await self.get(key)
            if isinstance(value, Bitarray):
                continue
            yield key, value

    async def expire(self, key: Key, timeout: float) -> int:
        return await self._run_in_executor(self._cache.touch, key, timeout)

    async def get_expire(self, key: Key) -> int:
        return await self._run_in_executor(self._get_expire, key)

    def _get_expire(self, key: Key) -> int:
        value, expire = self._cache.get(key, expire_time=True)
        if value is None:
            return NOT_EXIST
        if expire is None:
            return UNLIMITED
        return round((datetime.fromtimestamp(expire, timezone.utc) - datetime.now(timezone.utc)).total_seconds())

    async def get_size(self, key: Key) -> int:
        return -1

    async def ping(self, message: bytes | None = None) -> bytes:
        if message is None or message == b"PING":
            return b"PONG"
        return message

    async def clear(self):
        await self._run_in_executor(self._cache.clear)

    async def is_locked(
        self,
        key: Key,
        wait: float | None = None,
        step: float = 0.1,
    ) -> bool:
        if wait is None:
            return await self.exists(key)
        while wait > 0:
            if not await self.exists(key):
                return False
            wait -= step
            await asyncio.sleep(step)
        return await self.exists(key)

    async def unlock(self, key: Key, value: Value) -> bool:
        value = await self._serializer.encode(self, key=key, value=value, expire=None)
        return await self._run_in_executor(self._unlock, key, value)

    def _unlock(self, key: Key, value: Value) -> bool:
        if self._cache.get(key) == value:
            self._cache.delete(key)
            return True
        return False

    async def slice_incr(
        self,
        key: Key,
        start: int | float,
        end: int | float,
        maxvalue: int,
        expire: float | None = None,
    ) -> int:
        val_set = await self.get(key)
        count = 0
        new_val = []
        if val_set:
            for val in val_set:
                if start <= val <= end:
                    count += 1
                    new_val.append(val)

        if count < maxvalue:
            count += 1
            new_val.append(end)
        await self.set(key, new_val, expire=expire)
        return count

    async def set_add(self, key: Key, *values: str, expire: float | None = None):
        val = await self.get(key, default=set())
        val.update(values)
        await self.set(key, val, expire=expire)

    async def set_remove(self, key: Key, *values: str):
        val = await self.get(key, default=set())
        val.difference_update(values)
        await self.set(key, val)

    async def set_pop(self, key: Key, count: int = 100) -> Iterable[str]:
        values = await self.get(key, default=set())
        _values = []
        for _ in range(count):
            if not values:
                break
            _values.append(values.pop())

        await self.set(key, values)
        return _values

    async def get_keys_count(self) -> int:
        return await self._run_in_executor(lambda: len(self._cache))



from enum import Enum


class Command(Enum):
    GET = "get"
    GET_MANY = "get_many"
    GET_RAW = "get_raw"
    GET_MATCH = "get_match"
    SET = "set"
    SET_RAW = "set_raw"
    SET_MANY = "set_many"
    DELETE = "delete"
    DELETE_MANY = "delete_many"
    DELETE_MATCH = "delete_match"

    EXISTS = "exists"
    EXIST = "exists"
    SCAN = "scan"
    INCR = "incr"
    EXPIRE = "expire"
    GET_EXPIRE = "get_expire"
    CLEAR = "clear"

    SET_LOCK = "set_lock"
    UNLOCK = "unlock"
    IS_LOCKED = "is_locked"

    GET_BITS = "get_bits"
    INCR_BITS = "incr_bits"

    SLICE_INCR = "slice_incr"

    SET_ADD = "set_add"
    SET_REMOVE = "set_remove"
    SET_POP = "set_pop"

    PING = "ping"
    GET_SIZE = "get_size"
    GET_KEYS_COUNT = "get_keys_count"


ALL = set(Command)
PATTERN_CMDS = {Command.GET_MATCH, Command.DELETE_MATCH, Command.SCAN}
RETRIEVE_CMDS = {Command.GET, Command.INCR, Command.GET_MANY, Command.GET_MATCH}



from __future__ import annotations

from ._typing import AsyncCallable_T, Middleware, Result_T
from .backends.interface import Backend
from .commands import PATTERN_CMDS, Command
from .utils import get_obj_size


def add_prefix(prefix: str) -> Middleware:
    async def _middleware(call: AsyncCallable_T, cmd: Command, backend: Backend, *args, **kwargs) -> Result_T:
        if cmd in (Command.GET_MANY, Command.DELETE_MANY):
            return await call(*[prefix + key for key in args])
        if cmd == Command.SET_MANY:
            kwargs["pairs"] = {prefix + key: value for key, value in kwargs["pairs"].items()}
            return await call(**kwargs)

        as_key = "pattern" if cmd in PATTERN_CMDS else "key"
        key = kwargs.get(as_key)
        if key:
            kwargs[as_key] = prefix + key
            return await call(**kwargs)
        if args:
            key = args[0].lower()
            return await call(key, *args[1:], **kwargs)
        return await call(**kwargs)

    return _middleware


def all_keys_lower() -> Middleware:
    async def _middleware(call: AsyncCallable_T, cmd: Command, backend: Backend, *args, **kwargs) -> Result_T:
        if cmd in (Command.GET_MANY, Command.DELETE_MANY):
            return await call(*[key.lower() for key in args])

        if cmd == Command.SET_MANY:
            kwargs["pairs"] = {key.lower(): value for key, value in kwargs["pairs"].items()}
            return await call(**kwargs)

        as_key = "pattern" if cmd in PATTERN_CMDS else "key"

        key = kwargs.get(as_key)
        if key:
            kwargs[as_key] = key.lower()
            return await call(**kwargs)
        if args:
            key = args[0].lower()
            return await call(key, *args[1:], **kwargs)
        return await call(**kwargs)

    return _middleware


def memory_limit(min_bytes: int = 0, max_bytes: int | None = None) -> Middleware:
    async def _middleware(call: AsyncCallable_T, cmd: Command, backend: Backend, *args, **kwargs) -> Result_T | None:
        if cmd == Command.SET_MANY:
            pairs = {}
            for key, value in kwargs["pairs"].items():
                value_size = get_obj_size(value)
                if max_bytes and value_size > max_bytes or value_size < min_bytes:
                    continue
                pairs[key] = value
            if not pairs:
                return None
            kwargs["pairs"] = pairs
        elif cmd == Command.SET:
            value_size = get_obj_size(kwargs["value"])
            if max_bytes and value_size > max_bytes or value_size < min_bytes:
                return None
        return await call(*args, **kwargs)

    return _middleware



import asyncio
from unittest.mock import Mock

import pytest

from cashews import Cache, with_exceptions


async def test_iterator(cache: Cache):
    chunks = [b"a", b"b", b"c", b"d"]
    call = Mock()

    @cache.iterator(ttl=10, key="iterator")
    async def func():
        for chunk in chunks:
            call()
            await asyncio.sleep(0)
            yield chunk

    i = 0
    async for chunk in func():
        assert chunk == chunks[i]
        i += 1
    assert call.call_count == 4

    call.reset_mock()

    i = 0
    async for chunk in func():
        assert chunk == chunks[i]
        i += 1

    assert call.call_count == 0


class MyException(Exception):
    pass


async def test_iterator_error_with_cond(cache: Cache):
    call = Mock(side_effect=["a", "b", MyException(), "c", "d"])

    @cache.iterator(ttl=10, key="iterator", condition=with_exceptions(MyException))
    async def func():
        while True:
            try:
                yield call()
            except StopIteration:
                return
            await asyncio.sleep(0)

    full = ""
    with pytest.raises(MyException):
        async for chunk in func():
            assert chunk
            full += chunk

    assert full == "ab"
    assert call.call_count == 3
    call.reset_mock()

    with pytest.raises(MyException):
        async for chunk in func():
            assert chunk
            full += chunk

    assert full == "abab"
    assert call.call_count == 0


async def test_iterator_error(cache: Cache):
    call = Mock(side_effect=["a", MyException(), "c"])

    @cache.iterator(ttl=10, key="iterator")
    async def func():
        while True:
            try:
                yield call()
            except StopIteration:
                return
            await asyncio.sleep(0)

    with pytest.raises(MyException):
        async for chunk in func():
            assert chunk == "a"

    assert call.call_count == 2  # return a and raise error

    async for chunk in func():
        assert chunk == "c"

    assert call.call_count == 4  # return a, raise error + return c and stopIteration

    async for chunk in func():
        assert chunk == "c"

    assert call.call_count == 4



from __future__ import annotations

import logging
from datetime import datetime, timedelta, timezone
from functools import wraps
from typing import TYPE_CHECKING, Callable

from cashews.key import get_cache_key, get_cache_key_template
from cashews.ttl import ttl_to_seconds

from .defaults import _empty, context_cache_detect

if TYPE_CHECKING:  # pragma: no cover
    from cashews import Cache
    from cashews._typing import TTL, CallableCacheCondition, DecoratedFunc, KeyOrTemplate, Tags

__all__ = ("soft",)


logger = logging.getLogger(__name__)


def soft(
    backend: Cache,
    ttl: TTL,
    key: KeyOrTemplate | None = None,
    soft_ttl: TTL | None = None,
    exceptions: type[Exception] | tuple[type[Exception], ...] = Exception,
    condition: CallableCacheCondition = lambda *args, **kwargs: True,
    prefix: str = "soft",
    tags: Tags = (),
) -> Callable[[DecoratedFunc], DecoratedFunc]:
    """
    Cache strategy that allow to use pre-expiration

    :param backend: cache backend
    :param ttl: duration in seconds to store a result
    :param key: custom cache key, may contain alias to args or kwargs passed to a call
    :param exceptions: exceptions at which returned cache result if not reach ttl
    :param soft_ttl: pre expire ttl
    :param condition: callable object that determines whether the result will be saved or not
    :param prefix: custom prefix for key, default 'soft'
    :param tags: aliases for keys that used for cache (used for invalidation)
    """
    ttl = ttl_to_seconds(ttl)

    def _decor(func: DecoratedFunc) -> DecoratedFunc:
        _key_template = get_cache_key_template(func, key=key, prefix=prefix)
        for tag in tags:
            backend.register_tag(tag, _key_template)

        @wraps(func)
        async def _wrap(*args, **kwargs):
            _tags = [get_cache_key(func, tag, args, kwargs) for tag in tags]
            _cache_key = get_cache_key(func, _key_template, args, kwargs)
            cached = await backend.get(_cache_key, default=_empty)
            if cached is not _empty:
                soft_expire_at, result = cached
                if soft_expire_at > datetime.now(timezone.utc):
                    context_cache_detect._set(
                        _cache_key,
                        ttl=ttl,
                        soft_ttl=soft_ttl,
                        name="soft",
                        template=_key_template,
                        value=result,
                        soft_expire_at=soft_expire_at,
                    )
                    return result

            try:
                result = await func(*args, **kwargs)
            except exceptions:
                if cached is not _empty:
                    soft_expire_at, result = cached
                    context_cache_detect._set(
                        _cache_key,
                        ttl=ttl,
                        soft_ttl=soft_ttl,
                        name="soft",
                        template=_key_template,
                        value=result,
                        soft_expire_at=soft_expire_at,
                    )
                    return result
                raise
            else:
                if condition(result, args, kwargs, _cache_key):
                    _ttl = ttl_to_seconds(ttl, *args, result=result, **kwargs, with_callable=True)
                    _soft_ttl = ttl_to_seconds(soft_ttl, *args, result=result, **kwargs) or _ttl * 0.33
                    soft_expire_at = datetime.now(timezone.utc) + timedelta(seconds=_soft_ttl)
                    await backend.set(_cache_key, [soft_expire_at, result], expire=_ttl, tags=_tags)
                return result

        return _wrap  # type: ignore[return-value]

    return _decor



import asyncio
from unittest.mock import Mock

import pytest

from cashews import decorators
from cashews.backends.memory import Memory


async def test_lock_cache_parallel(cache):
    mock = Mock()

    @cache.locked(key="key", wait=False, ttl=10)
    async def func():
        await asyncio.sleep(0.1)
        mock()

    for _ in range(2):
        await asyncio.gather(*[func() for _ in range(10)], return_exceptions=True)

    assert mock.call_count == 2


async def test_lock_cache_parallel_with_ttl(cache):
    mock = Mock()

    @cache.locked(key="key", ttl=1, wait=True)
    async def func():
        await asyncio.sleep(0.01)
        mock()

    for _ in range(2):
        await asyncio.gather(*[func() for _ in range(10)])

    assert mock.call_count == 20


async def test_lock_cache_parallel_check_interval(cache):
    mock = Mock()
    mock_middleware = Mock()

    async def middleware(call, cmd, backend, *args, **kwargs):
        mock_middleware()
        return await call(*args, **kwargs)

    cache.add_middleware(middleware)

    @cache.locked(key="key", ttl=1, wait=True, check_interval=0.05)
    async def func_interval():
        await asyncio.sleep(0.01)
        mock()

    for _ in range(2):
        await asyncio.gather(*[func_interval() for _ in range(2)])

    assert mock.call_count == 4
    assert mock_middleware.call_count == 12


async def test_lock_cache_iterator(cache):
    mock = Mock()
    chunks = range(10)

    @cache.locked(key="key", ttl=10)
    async def func():
        for chunk in chunks:
            mock()
            await asyncio.sleep(0)
            yield chunk

    async for _ in func():
        pass
    assert mock.call_count == 10


async def test_lock_cache_broken_backend():
    class BrokenMemoryBackend(Memory):
        async def ping(self, message=None) -> bytes:
            raise Exception("broken")

    backend = Mock(wraps=BrokenMemoryBackend())

    @decorators.locked(backend, key="key")
    async def func(resp=b"ok"):
        await asyncio.sleep(0.01)
        return resp

    with pytest.raises(Exception):
        await asyncio.gather(*[func() for _ in range(10)])


async def test_thunder_protection():
    mock = Mock()

    @decorators.thunder_protection(key="key")
    async def func(resp=b"ok"):
        await asyncio.sleep(0.01)
        mock()
        return resp

    for _ in range(2):
        await asyncio.gather(*[func() for _ in range(10)])

    assert mock.call_count == 2



from __future__ import annotations

import asyncio
from functools import wraps
from typing import TYPE_CHECKING, Callable

from cashews.key import get_cache_key, get_cache_key_template
from cashews.ttl import ttl_to_seconds

from ._exception import RaiseException, return_or_raise
from .defaults import _empty, context_cache_detect

if TYPE_CHECKING:  # pragma: no cover
    from cashews import Cache
    from cashews._typing import TTL, CallableCacheCondition, DecoratedFunc, Key, KeyOrTemplate, Tags

__all__ = ("hit",)


def hit(
    backend: Cache,
    ttl: TTL,
    cache_hits: int,
    update_after: int | None = None,
    key: KeyOrTemplate | None = None,
    condition: CallableCacheCondition = lambda *args, **kwargs: True,
    prefix: str = "hit",
    tags: Tags = (),
    background: bool = True,
) -> Callable[[DecoratedFunc], DecoratedFunc]:
    """
    Cache call results and drop cache after given numbers of call 'cache_hits'
    :param backend: cache backend
    :param ttl: duration in seconds to store a result
    :param cache_hits: number of cache hits till cache will dropped
    :param update_after: number of cache hits after cache will update
    :param key: custom cache key, may contain alias to args or kwargs passed to a call
    :param condition: callable object that determines whether the result will be saved or not
    :param prefix: custom prefix for key, default 'hit'
    :param tags: aliases for keys that used for cache (used for invalidation)
    :param background: if true will run recalculation in background
    """
    ttl = ttl_to_seconds(ttl)
    background_tasks = set()

    def _decor(func: DecoratedFunc) -> DecoratedFunc:
        _key_template = get_cache_key_template(func, key=key, prefix=prefix)
        for tag in tags:
            backend.register_tag(tag, _key_template + ":counter")
            backend.register_tag(tag, _key_template)

        @wraps(func)
        async def _wrap(*args, **kwargs):
            _cache_key = get_cache_key(func, _key_template, args, kwargs)
            _tags = [get_cache_key(func, tag, args, kwargs) for tag in tags]

            call_args = (func, args, kwargs, backend, _cache_key, ttl, condition, _tags)

            cached, hits = await asyncio.gather(
                backend.get(_cache_key, default=_empty),
                backend.incr(_cache_key + ":counter", expire=ttl, tags=_tags),
            )
            if cached is not _empty and hits and hits <= cache_hits:
                _ttl = ttl_to_seconds(ttl, *args, **kwargs, result=cached, with_callable=True)

                context_cache_detect._set(
                    _cache_key,
                    ttl=_ttl,
                    cache_hits=cache_hits,
                    name="hit",
                    template=_key_template,
                    value=cached,
                )
                if update_after and hits == update_after:
                    task = asyncio.create_task(_get_and_save(*call_args))
                    background_tasks.add(task)
                    task.add_done_callback(background_tasks.discard)
                    if not background:
                        await task
                return return_or_raise(cached)
            return await _get_and_save(*call_args)

        return _wrap  # type: ignore[return-value]

    return _decor


async def _get_and_save(func, args, kwargs, backend: Cache, key: Key, ttl, store, tags):
    _exc = None
    try:
        result = await func(*args, **kwargs)
    except Exception as exc:
        _exc = exc
        result = exc

    ttl = ttl_to_seconds(ttl, *args, result=result, **kwargs, with_callable=True)
    cond_result = store(result, args, kwargs, key=key)
    to_cache = None
    if isinstance(cond_result, bool) and cond_result and not isinstance(result, Exception):
        to_cache = (result,)
    elif isinstance(cond_result, Exception):
        to_cache = (RaiseException(result),)

    if to_cache is not None:
        await asyncio.gather(
            backend.delete(key + ":counter"),
            backend.set(key, to_cache[0], expire=ttl, tags=tags),
        )
    if _exc:
        raise _exc
    return result



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Callable, Union
from urllib.parse import parse_qsl, urlparse

from cashews.backends.interface import Backend
from cashews.backends.memory import Memory
from cashews.exceptions import BackendNotAvailableError
from cashews.picklers import PicklerType

if TYPE_CHECKING:  # pragma: no cover
    BackendOrFabric = Union[type[Backend], Callable[..., Backend]]

_NO_REDIS_ERROR = "Redis backend requires `redis` to be installed."
_CUSTOM_ERRORS = {
    "redis": _NO_REDIS_ERROR,
    "rediss": _NO_REDIS_ERROR,
    "disk": "Disk backend requires `diskcache` to be installed.",
}
_BACKENDS: dict[str, tuple[BackendOrFabric, bool, PicklerType]] = {}


def register_backend(
    alias: str,
    backend_class: BackendOrFabric,
    pass_uri: bool = False,
    pickler: PicklerType = PicklerType.NULL,
) -> None:
    _BACKENDS[alias] = (backend_class, pass_uri, pickler)


register_backend("mem", Memory)


try:
    import redis  # noqa: F401
except ImportError:
    pass
else:
    from cashews.backends.redis import Redis
    from cashews.backends.redis.client_side import BcastClientSide

    def _redis_fabric(**params) -> Redis | BcastClientSide:
        if params.pop("client_side", None):
            return BcastClientSide(**params)
        return Redis(**params)

    register_backend("redis", _redis_fabric, pass_uri=True, pickler=PicklerType.DEFAULT)
    register_backend("rediss", _redis_fabric, pass_uri=True, pickler=PicklerType.DEFAULT)


try:
    import diskcache  # noqa: F401
except ImportError:
    pass
else:
    from cashews.backends.diskcache import DiskCache

    register_backend("disk", DiskCache, pickler=PicklerType.DEFAULT)


def settings_url_parse(url: str) -> tuple[BackendOrFabric, dict[str, Any], PicklerType]:
    parse_result = urlparse(url)
    params: dict[str, Any] = dict(parse_qsl(parse_result.query))
    params = _serialize_params(params)

    alias = parse_result.scheme
    if alias == "":
        return Memory, {"disable": True}, PicklerType.NULL

    if alias not in _BACKENDS:
        error = _CUSTOM_ERRORS.get(alias, f"wrong backend alias {alias}")
        raise BackendNotAvailableError(error)
    backend_class, pass_uri, pickler = _BACKENDS[alias]
    if pass_uri:
        params["address"] = url.split("?")[0]
    return backend_class, params, pickler


def _serialize_params(params: dict[str, str]) -> dict[str, str | int | bool | float]:
    new_params = {}
    bool_keys = ("safe", "suppress", "enable", "disable", "client_side")
    true_values = (
        "1",
        "true",
    )
    for key, value in params.items():
        _value: str | int | bool | float
        if key.lower() in bool_keys:
            _value = value.lower() in true_values
        elif value.isdigit():
            _value = int(value)
        else:
            try:
                _value = float(value)
            except ValueError:
                _value = value
        new_params[key.lower()] = _value
    return new_params



from __future__ import annotations

import asyncio
import math
from collections import namedtuple
from collections.abc import Iterable
from functools import wraps
from typing import TYPE_CHECKING, Any, Callable, Union

from cashews.backends.interface import Backend
from cashews.key import get_cache_key, get_cache_key_template
from cashews.utils import get_indexes

if TYPE_CHECKING:  # pragma: no cover
    from cashews._typing import DecoratedFunc, KeyOrTemplate


__all__ = ("bloom",)

TrueFalsePair = tuple[int, int]
IntOrPair = Union[int, TrueFalsePair]
BloomParams = namedtuple("BloomParams", ("size", "number_of_buckets"))

possible_set = all


def all_zeros(values: Iterable[int]) -> bool:
    return all(v == 0 for v in values)


def not_set(values: Iterable[int]) -> bool:
    return not all(values)


def bloom(
    backend: Backend,
    *,
    capacity: int,
    name: KeyOrTemplate | None = None,
    false_positives: float | int = 1,
    check_false_positive: bool = True,
    prefix: str = "bloom",
) -> Callable[[DecoratedFunc], DecoratedFunc]:
    """
    Decorator that can help you to use bloom filter algorithm

    @bloom(name="user_name:{name}", false_positives=1, capacity=10_000)
    async def is_user_exists(name) -> bool:
        return name in ....

    :param backend: cache backend
    :param name: custom cache key
    :param capacity: the same as n - number of elements
    :param false_positives: Percents of false positive results
    :param check_false_positive: do we need to check if we have positive result
    :param prefix: custom prefix for key, default 'bloom'
    """
    assert false_positives and capacity
    assert 0 < false_positives < 100
    index_size, number_of_buckets = params_for(capacity, false_positives / 100)

    def _decor(func: DecoratedFunc) -> DecoratedFunc:
        _name = get_cache_key_template(func, key=name)
        _cache_key = f"{_name}:{index_size}"
        if prefix:
            _cache_key = f"{prefix}:{_cache_key}"

        _set = getattr(func, "set", None)

        async def __set(*args: Any, **kwargs: Any):
            if _set is None:
                result = await func(*args, **kwargs)
            else:
                result = await _set(*args, **kwargs)
            if not result:
                return result
            _bloom_key = get_cache_key(func, _name, args, kwargs)
            indexes = get_indexes(_bloom_key, number_of_buckets, index_size)
            await backend.incr_bits(_cache_key, *indexes)
            return result

        func.set = __set  # type: ignore[attr-defined]

        @wraps(func)
        async def _wrap(*args, **kwargs):
            _bloom_key = get_cache_key(func, _name, args, kwargs)
            hashes = get_indexes(_bloom_key, number_of_buckets, index_size)
            values = await backend.get_bits(_cache_key, *hashes)
            if values is None:
                return await func(*args, **kwargs)
            if possible_set(values):  # if all bits is set
                # false positive
                if check_false_positive:
                    return await func(*args, **kwargs)
                return True
            return False

        return _wrap  # type: ignore[return-value]

    return _decor


def dual_bloom(
    backend: Backend,
    *,
    capacity: IntOrPair,
    name: KeyOrTemplate | None = None,
    false: IntOrPair = 1,
    no_collisions: bool = False,
    prefix: str = "dual_bloom",
) -> Callable[[DecoratedFunc], DecoratedFunc]:
    """
    Decorator that can help you to use bloom filter algorithm
     this is  implementation with 2 bloom filters - one for true and another for false
     That give the ability use bloom filter without pre-filling data
     but with possible false positive and false negative results

    @dual_bloom(name="user_name:{name}", false=1, capacity=10_000)
    async def is_user_exists(name) -> bool:
        return name in ....

    :param backend: cache backend
    :param capacity: the same as n - number of elements
    :param name: custom cache key
    :param no_collisions: add value only no collisions
    :param false: Percents of false results
    :param prefix: custom prefix for key, default 'dual_bloom'
    """
    filters_params = _get_params_for_filters(false, capacity)

    def _decor(func: DecoratedFunc) -> DecoratedFunc:
        _cache_key = get_cache_key_template(func, key=name)
        if prefix:
            _cache_key = f"{prefix}:{_cache_key}"
        _true_bloom_key = _cache_key + ":true"
        _false_bloom_key = _cache_key + ":false"

        @wraps(func)
        async def _wrap(*args, **kwargs):
            _bloom_key = get_cache_key(func, _cache_key, args, kwargs)
            indexes_true, indexes_false = _get_indexes(_bloom_key, *filters_params)

            true_values, false_values = await asyncio.gather(
                backend.get_bits(_true_bloom_key, *indexes_true),
                backend.get_bits(_false_bloom_key, *indexes_false),
            )
            if not_set(true_values) and not_set(false_values):
                # not set yet
                result = await func(*args, **kwargs)
                if result and (not no_collisions or all_zeros(true_values)):
                    await backend.incr_bits(_true_bloom_key, *indexes_true)
                if not result and (not no_collisions or all_zeros(false_values)):
                    await backend.incr_bits(_false_bloom_key, *indexes_false)
                return result
            if not_set(true_values) and possible_set(false_values):
                return False  # can be false Negative
            if not_set(false_values) and possible_set(true_values):
                return True  # can be false Positive
            return await func(*args, **kwargs)

        return _wrap  # type: ignore[return-value]

    return _decor


def _get_params_for_filters(
    false: IntOrPair = 1,
    capacity: IntOrPair | None = None,
) -> tuple[BloomParams, BloomParams]:
    assert false and capacity
    capacity_true, capacity_false = _to_pair(capacity)
    false_true, false_false = _to_pair(false)
    assert 0 < false_true < 100
    assert 0 < false_false < 100
    return params_for(capacity_true, false_true / 100), params_for(capacity_false, false_false / 100)


def _to_pair(value: IntOrPair) -> TrueFalsePair:
    if isinstance(value, int):
        return value, value
    return value


def _get_indexes(key: str, params_true: BloomParams, params_false: BloomParams):
    index_size_true, number_of_buckets_true = params_true
    index_size_false, number_of_buckets_false = params_false
    return (
        get_indexes(key + "true", number_of_buckets_true, index_size_true),
        get_indexes(key + "false", number_of_buckets_false, index_size_false),
    )


def params_for(capacity: int, false_positives: float = 0.01) -> BloomParams:
    """
    capacity = n
    k=(m * ln 2 / n)
    m = k * n / ln 2
    p = (1-e^{-k * n/m})^{k}
    return m - number of bits in the array and k the number of hash functions
    """
    m = _count_m(capacity, false_positives)
    k = _count_k(m, capacity)
    assert k > 0, "too high false positive value"
    return BloomParams(m, k)


def _count_k(m, n):
    return int(round((m / n) * math.log(2)))


def _count_k_from_p(p):
    return int(math.ceil(math.log(1.0 / p, 2)))


def _count_m(n, p=0.1):
    return int(math.ceil(-1 * n * math.log(p) / math.log(2) ** 2))


def _count_probability(n, m, k):
    ev = -(k * n / m)
    ome = 1 - math.e**ev
    return math.pow(ome, k)



import asyncio
import random

import pytest

from cashews import LockedError
from cashews.backends.interface import NOT_EXIST, UNLIMITED
from cashews.wrapper import Cache
from cashews.wrapper.transaction import TransactionMode


@pytest.fixture(
    name="tx_mode",
    params=[
        TransactionMode.FAST,
        TransactionMode.LOCKED,
        TransactionMode.SERIALIZABLE,
    ],
)
def _tx_mode(request):
    return request.param


async def test_transaction_set(cache: Cache, tx_mode):
    await cache.set("key1", "value1", expire=1)
    await cache.set("key2", "value2", expire=2)

    async with cache.transaction(tx_mode):
        await cache.set("key1", "value2", expire=3)
        await cache.set("key", "value")

        assert await cache.get("key") == "value"
        assert await cache.get("key1") == "value2"
        assert await cache.get("key2") == "value2"

    assert await cache.get("key") == "value"
    assert await cache.get("key1") == "value2"
    assert await cache.get("key2") == "value2"


async def test_transaction_set_rollback(cache: Cache, tx_mode):
    await cache.set("key1", "value1", expire=1)
    await cache.set("key2", "value2", expire=2)

    async with cache.transaction(tx_mode) as tx:
        await cache.set("key1", "value2", expire=3)
        await cache.set("key", "value")

        assert await cache.get("key") == "value"
        assert await cache.get("key1") == "value2"
        assert await cache.get("key2") == "value2"
        await tx.rollback()

        assert await cache.get("key") is None
        assert await cache.get("key1") == "value1"
        assert await cache.get("key2") == "value2"

    assert await cache.get("key") is None
    assert await cache.get("key1") == "value1"
    assert await cache.get("key2") == "value2"


async def test_transaction_set_exception(cache: Cache, tx_mode):
    await cache.set("key1", "value1", expire=1)
    await cache.set("key2", "value2", expire=2)

    with pytest.raises(ValueError):
        async with cache.transaction(tx_mode):
            await cache.set("key1", "value2", expire=3)
            await cache.set("key", "value")

            assert await cache.get("key") == "value"
            assert await cache.get("key1") == "value2"
            assert await cache.get("key2") == "value2"
            raise ValueError("rollback")

    assert await cache.get("key") is None
    assert await cache.get("key1") == "value1"
    assert await cache.get("key2") == "value2"


async def test_transaction_get_or_set(cache: Cache, tx_mode):
    await cache.set("key1", "value1", expire=1)

    async with cache.transaction(tx_mode):
        await cache.get_or_set("key", "value")
        await cache.get_or_set("key1", "value2")

        assert await cache.get("key") == "value"
        assert await cache.get("key1") == "value1"

    assert await cache.get("key") == "value"
    assert await cache.get("key1") == "value1"


async def test_transaction_get_or_set_rollback(cache: Cache, tx_mode):
    await cache.set("key1", "value1", expire=1)

    async with cache.transaction(tx_mode) as tx:
        await cache.get_or_set("key", "value")
        await cache.get_or_set("key1", "value2")

        assert await cache.get("key") == "value"
        assert await cache.get("key1") == "value1"

        await tx.rollback()

        assert await cache.get("key") is None
        assert await cache.get("key1") == "value1"

    assert await cache.get("key") is None
    assert await cache.get("key1") == "value1"


async def test_transaction_set_delete_get_many(cache: Cache, tx_mode):
    await cache.set("key1", "value", expire=1)
    await cache.set("key2", "value2", expire=2)

    async with cache.transaction(tx_mode):
        await cache.delete("key1")
        assert await cache.get("key1") is None

        await cache.set_many({"key1": "value1", "key3": "value3"}, expire=3)
        await cache.delete("key")
        await cache.delete_many("key2", "key3")

        assert await cache.get("key") is None
        assert await cache.get("key1") == "value1"
        assert await cache.get("key2") is None
        assert await cache.get("key3") is None

        await cache.set("key2", "value2")
        assert await cache.get_many("key", "key1", "key2") == (None, "value1", "value2")

    assert await cache.get("key") is None
    assert await cache.get("key1") == "value1"
    assert await cache.get("key2") == "value2"
    assert await cache.get("key3") is None
    assert await cache.get_many("key", "key1", "key2") == (None, "value1", "value2")


async def test_transaction_set_delete_match_get_many_match(cache: Cache, tx_mode):
    await cache.set("kek", "value")
    await cache.set("key1", "value", expire=1)
    await cache.set("key2", "value2", expire=2)

    async with cache.transaction(tx_mode):
        await cache.set("key1", "value1", expire=3)
        await cache.set("key3", "value3")
        await cache.delete_match("key*")

        assert await cache.get("key1") is None
        assert await cache.get("key2") is None
        assert await cache.get("key3") is None

        await cache.set("key2", "value2")
        assert await cache.get_many("kek", "key1", "key2") == ("value", None, "value2")
        match = sorted([(key, value) async for key, value in cache.get_match("k*")])
        assert len(match) == 2
        assert match[0] == ("kek", "value")
        assert match[1] == ("key2", "value2")

    assert await cache.get("key1") is None
    assert await cache.get("key2") == "value2"
    assert await cache.get("key3") is None

    assert await cache.get_many("kek", "key1", "key2") == ("value", None, "value2")


async def test_transaction_incr(cache: Cache, tx_mode):
    await cache.incr("key1")
    await cache.incr("key3")
    await cache.incr("key4")

    async with cache.transaction(tx_mode):
        assert await cache.incr("key1") == 2
        assert await cache.incr("key1") == 3

        assert await cache.incr("key2") == 1

        assert await cache.delete("key3")
        assert await cache.incr("key3") == 1

        assert await cache.incr("key4") == 2
        assert await cache.delete("key4")

    assert await cache.incr("key1") == 4
    assert await cache.incr("key2") == 2
    assert await cache.incr("key3") == 2
    assert await cache.incr("key4") == 1


async def test_transaction_expire(cache: Cache, tx_mode):
    await cache.set("key1", "value1")
    await cache.set("key3", "value3")
    await cache.set("key4", "value4")

    async with cache.transaction(tx_mode):
        await cache.delete("key3")
        await cache.set("key2", "value2", expire=100)
        await cache.expire("key1", 11)
        await cache.expire("key2", 12)
        await cache.expire("key3", 13)

        assert await cache.get_expire("key1") >= 10
        assert await cache.get_expire("key2") >= 11
        assert await cache.get_expire("key3") < 0
        assert await cache.get_expire("key4") == UNLIMITED
        assert await cache.get_expire("key5") == NOT_EXIST

    assert await cache.get_expire("key1") >= 10
    assert await cache.get_expire("key2") >= 11
    assert await cache.get_expire("key3") < 0


async def test_transaction_exist(cache: Cache, tx_mode):
    await cache.set("key1", "value1")
    await cache.set("key3", "value3")

    async with cache.transaction(tx_mode):
        await cache.delete("key3")
        await cache.set("key2", "value2")

        assert not await cache.exists("key3")
        assert await cache.exists("key1")
        assert await cache.exists("key2")

    assert await cache.exists("key1")
    assert await cache.exists("key2")
    assert not await cache.exists("key3")


async def test_transaction_scan(cache: Cache, tx_mode):
    await cache.set("key1", "value1")
    await cache.set("key3", "value3")

    async with cache.transaction(tx_mode):
        await cache.delete("key3")
        await cache.set("key2", "value2")

        assert sorted([key async for key in cache.scan("k*")]) == ["key1", "key2"]

    assert sorted([key async for key in cache.scan("k*")]) == ["key1", "key2"]


async def test_isolation(cache: Cache, tx_mode):
    await cache.set("key1", "value1")
    await cache.set("key2", "value2")

    @cache.transaction(mode=tx_mode, timeout=1)
    async def do(value: str):
        assert await cache.get("key2") == "value2"  # no locks
        await cache.incr("incr")
        await cache.set("key1", "value")
        if value == "v4":
            raise Exception("rollback")
        await cache.set("key2", value)
        await asyncio.sleep(random.randint(1, 5) / 100)  # like race condition
        await cache.incr("incr")
        await cache.set("key3", value, exist=True)
        await cache.set("key1", value)

    await asyncio.gather(
        do("v1"),
        do("v2"),
        do("v3"),
        do("v4"),
        return_exceptions=True,
    )

    if tx_mode == TransactionMode.FAST:
        assert await cache.get("incr") == 2
    else:
        assert await cache.get("incr") == 6

    first_value = None
    async for _key, value in cache.get_match("k*"):
        if first_value is None:
            first_value = value
        else:
            assert value == first_value


async def test_inner_transaction_commit_outer_rollback(cache: Cache, tx_mode):
    await cache.set("key1", "value1")
    await cache.set("key2", "value2")

    async with cache.transaction(tx_mode) as tx:
        await cache.delete("key2")
        await cache.set("key1", "value2")
        async with cache.transaction(tx_mode):
            await cache.set("key1", "value3")
        await tx.rollback()

    assert await cache.get("key1") == "value1"
    assert await cache.get("key2") == "value2"


async def test_gather(cache: Cache, tx_mode):
    await cache.set("key1", "value1")
    await cache.set("key2", "value1")

    async with cache.transaction(tx_mode):
        await asyncio.gather(
            cache.set("key1", "value3"),
            cache.set("key1", "value4"),
            cache.delete("key2"),
            cache.incr("key3"),
        )

    assert await cache.get("key1") in ("value4", "value3")
    assert await cache.get("key2") is None
    assert await cache.get("key3") == 1


async def test_decorators_smoke(cache: Cache, tx_mode):
    @cache.transaction(tx_mode)
    @cache.failover(ttl=10)
    @cache.cache(ttl=10)
    @cache.soft(ttl=10)
    @cache.early(ttl=10, early_ttl=5)
    @cache.hit(ttl=10, cache_hits=2)
    @cache.invalidate("key")
    @cache.slice_rate_limit(limit=10, period=10)
    @cache.circuit_breaker(errors_rate=10, period=10, ttl=10)
    async def do_sum(*args):
        return sum(args)

    assert await do_sum(1, 2, 5) == 8
    assert await do_sum(1, 2, 5) == 8


async def test_long_running_transaction(cache):
    @cache.transaction(TransactionMode.LOCKED, timeout=0.1)
    async def func():
        await cache.set("key", "value")
        await asyncio.sleep(1)

    with pytest.raises(LockedError):
        await asyncio.gather(func(), func())
